import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import pypika
from pypika import AliasedQuery
from pypika import Field
from pypika import Interval
from pypika import Order
from pypika import Table
from pypika import analytics
from pypika.dialects import MySQLQuery
from pypika.dialects import PostgreSQLQuery
from pypika.dialects import SnowflakeQuery
from pypika.functions import Cast
from pypika.functions import DateAdd
from pypika.functions import Floor
from pypika.queries import Query
from pypika.queries import Selectable
from pypika.terms import AnalyticFunction
from pypika.terms import Function
from pypika.terms import LiteralValue
from pypika.terms import Term
from pypika.terms import Tuple as TupleTerm
from pypika.terms import WindowFrameAnalyticFunction
from pypika.utils import builder
from pypika.utils import format_alias_sql

from tecton_core.compute_mode import ComputeMode
from tecton_core.compute_mode import offline_retrieval_compute_mode
from tecton_core.data_types import DataType
from tecton_core.query.dialect import Dialect
from tecton_proto.data.feature_store__client_pb2 import FeatureStoreFormatVersion


class CaseSensitiveSnowflakeQueryBuilder(pypika.dialects.SnowflakeQueryBuilder):
    QUOTE_CHAR = '"'


class CaseSensitiveSnowflakeQuery(pypika.queries.Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> CaseSensitiveSnowflakeQueryBuilder:
        return CaseSensitiveSnowflakeQueryBuilder(**kwargs)


class DuckDBQuery(pypika.queries.Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "DuckDBQueryBuilder":
        return DuckDBQueryBuilder(**kwargs)

    @classmethod
    def with_(
        cls, table: Union[str, Selectable], name: str, materialized: Optional[bool] = False, **kwargs: Any
    ) -> "DuckDBQueryBuilder":
        return cls._builder(**kwargs).with_(table, name, materialized)


class DuckDBAliasedQuery(pypika.queries.AliasedQuery):
    def __init__(self, name: str, query: Optional[Selectable] = None, materialized: Optional[bool] = False) -> None:
        super().__init__(name, query)
        self.materialized = materialized


class DuckDBQueryBuilder(pypika.dialects.PostgreSQLQueryBuilder):
    ALIAS_QUOTE_CHAR = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(as_keyword=True, **kwargs)

    @builder
    def with_(self, selectable: Selectable, name: str, materialized: Optional[bool] = False) -> "DuckDBQueryBuilder":
        t = DuckDBAliasedQuery(name, selectable, materialized)
        self._with.append(t)

    def _with_sql(self, **kwargs: Any) -> str:
        return "WITH " + ",".join(
            clause.name
            + f" AS {'MATERIALIZED' if clause.materialized  else ''} ("
            + clause.get_sql(subquery=False, with_alias=False, **kwargs)
            + ") "
            for clause in self._with
        )


class CaseInsensitiveSnowflakeQueryBuilder(pypika.dialects.SnowflakeQueryBuilder):
    QUOTE_CHAR = None
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = ""
    QUERY_CLS = SnowflakeQuery


class CaseInsensitiveSnowflakeQuery(pypika.queries.Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> CaseInsensitiveSnowflakeQueryBuilder:
        return CaseInsensitiveSnowflakeQueryBuilder(**kwargs)


class CustomQuery(pypika.queries.QueryBuilder):
    """
    Defines a custom-query class. It's needed for us to wrap some user-defined sql string from transformations in a QueryBuilder object.
    """

    def __init__(self, sql: str) -> None:
        super().__init__()
        self.sql = sql
        self.withs: List[Selectable] = []

    def with_(self, selectable: Selectable, name: str) -> "CustomQuery":
        """
        overrides QueryBuilder.with_
        """
        t = AliasedQuery(name, selectable)
        self.withs.append(t)
        return self

    def get_sql(self, with_alias: bool = False, subquery: bool = False, **kwargs: Any) -> str:
        """
        overrides QueryBuilder.get_sql
        """
        sql = ""
        if self.withs:
            sql += "WITH " + ",".join(
                clause.name + " AS (" + clause.get_sql(subquery=False, with_alias=False, **kwargs) + ") "
                for clause in self.withs
            )
        sql += self.sql
        if with_alias:
            self.alias = "sq0"
            sql = f"({sql})"
            return format_alias_sql(sql, self.alias or self._table_name, **kwargs)
        return sql


class LastValue(analytics.LastValue):
    """
    Fixed version of pypika's LastValue to handle Snowflake and Athena wanting "ignore nulls"
     to be outside the parens of the window func, and Spark using a bool param.
    """

    def __init__(self, dialect: "Dialect", *args: Any, **kwargs: Any) -> None:
        self._dialect = dialect
        super().__init__(*args, **kwargs)

    def get_special_params_sql(self, **kwargs: Any) -> Optional[str]:
        if self._dialect == Dialect.SPARK:
            # see: https://docs.databricks.com/sql/language-manual/functions/last_value.html
            # for sparkSQL syntax
            return f", {self._ignore_nulls}"
        elif self._dialect == Dialect.DUCKDB and self._ignore_nulls:
            # see: https://duckdb.org/docs/sql/window_functions.html
            # for DuckDB syntax
            return " IGNORE NULLS"
        # Snowflake does not support adding a function param to indicate "ignore nulls"
        # It looks like LAST_VALUE(...) IGNORE NULLS OVER(...)
        # see: https://docs.snowflake.com/en/sql-reference/functions/last_value.html
        # Nor does Athena - Athena docs don't make this clear though.
        else:
            return None

    def get_function_sql(self, **kwargs: Any) -> str:
        if self._dialect == Dialect.SPARK or self._dialect == Dialect.DUCKDB:
            return super(LastValue, self).get_function_sql(**kwargs)
        else:
            function_sql = super(AnalyticFunction, self).get_function_sql(**kwargs)
            partition_sql = self.get_partition_sql(**kwargs)

            sql = function_sql
            if self._ignore_nulls:
                sql += " IGNORE NULLS"
            if self._include_over:
                sql += " OVER({partition_sql})".format(partition_sql=partition_sql)
            return sql


class CompatFunctions:
    def __init__(self, compute_mode: ComputeMode) -> None:
        self.compute_mode = compute_mode

    @staticmethod
    def for_dialect(d: Dialect, compute_mode: Optional[ComputeMode] = None) -> "CompatFunctions":
        compute_mode = offline_retrieval_compute_mode(compute_mode)
        if d == Dialect.SPARK:
            return _Spark(compute_mode=compute_mode)
        elif d == Dialect.DUCKDB:
            return _DuckDB(compute_mode=compute_mode)
        elif d == Dialect.SNOWFLAKE:
            return _Snowflake(compute_mode=compute_mode)
        elif d == Dialect.ATHENA:
            return _Athena(compute_mode=compute_mode)
        elif d == Dialect.TORCH:
            return _Torch(compute_mode=compute_mode)
        elif d == Dialect.BIGQUERY:
            return _Bigquery(compute_mode=compute_mode)
        msg = f"Unexpected dialect {d}"
        raise Exception(msg)

    @classmethod
    def query(cls) -> typing.Type[Query]:
        raise NotImplementedError()

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        raise NotImplementedError()

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        raise NotImplementedError()

    @classmethod
    def ordered_filtered_list(
        cls, from_column: Term, order_by_column: Term, filter_clause: Term, direction: Order = Order.asc
    ) -> Term:
        raise NotImplementedError()

    @classmethod
    def list(cls, column: str) -> WindowFrameAnalyticFunction:
        raise NotImplementedError()

    @classmethod
    def any(cls, column: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def list_filter_nulls(cls, column: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def list_transform(cls, column: Term, lambda_func: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def to_timestamp(cls, time_str: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        raise NotImplementedError()

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        raise NotImplementedError()

    @classmethod
    def to_utc(cls, timestamp: Term) -> Term:
        raise NotImplementedError()

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        raise NotImplementedError()

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        raise NotImplementedError()

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        raise NotImplementedError()

    @classmethod
    def convert_epoch_term_in_seconds(
        cls, timestamp: Term, time_stamp_feature_store_format_version: FeatureStoreFormatVersion
    ) -> Term:
        """
        Converts an epoch term to a timestamp column
        :param timestamp: epoch column [V0 : Seconds, V1 : Nanoseconds]
        :param time_stamp_feature_store_format_version: Feature Store Format Version
        :return: epoch in seconds
        """
        if time_stamp_feature_store_format_version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
            return timestamp
        # Cast explicitly to an integer (which is the expectation), since otherwise SQL will often treat integer
        # division as outputting a double, which is incompatible with to_timestamp(epoch_seconds) functions
        return Cast(timestamp / int(1e9), as_type="bigint")

    @classmethod
    def convert_epoch_seconds_to_feature_store_format_version(
        cls, timestamp: Term, feature_store_format_version: FeatureStoreFormatVersion
    ) -> Term:
        if feature_store_format_version == FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
            return timestamp
        return timestamp * int(1e9)


class _Snowflake(CompatFunctions):
    def query(self) -> typing.Type[Query]:
        # Compute.SNOWFLAKE is "legacy" snowflake, executing snowflake queries in unified Tecton mode will have
        # ComputeMode.DUCK_DB
        return (
            CaseInsensitiveSnowflakeQuery if self.compute_mode == ComputeMode.SNOWFLAKE else CaseSensitiveSnowflakeQuery
        )

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        return Function("array_construct", *[Field(n) for n in field_names])

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        res = []
        for idx in range(len(aliases)):
            val = Function("get", Field(name), idx)
            if field_names[idx] in schema:
                feature_type = schema[field_names[idx]].sql_type
                val = Cast(val, feature_type)
            res.append(val.as_(aliases[idx]))
        return res

    @classmethod
    def to_timestamp(cls, time_str: str) -> Term:
        return Function("to_timestamp", time_str)

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        # snowflake uses dateadd rather than date_add;
        # https://docs.snowflake.com/en/sql-reference/functions/dateadd.html

        # LiteralValue will not put quotes around.
        # So we get dateadd(second, ...)
        interval = LiteralValue(interval)
        return Function("dateadd", interval, amount, time_field)

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        return Function("date_part", Field("epoch_second"), timestamp)

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        return Function("to_timestamp", unix_timestamp)

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        raise NotImplementedError()

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        raise NotImplementedError()


class DuckDBList(WindowFrameAnalyticFunction):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super(DuckDBList, self).__init__("LIST", *args, **kwargs)


class DuckDBOrderedFilteredList(Term):
    def __init__(
        self,
        from_column: Term,
        order_by_column: Term,
        filter_clause: Term,
        direction: Order = Order.asc,
    ) -> None:
        super().__init__()
        self.from_column = from_column
        self.order_by_column = order_by_column
        self.direction = direction
        self.filter_clause = filter_clause

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns a `LIST()` with an `ORDER BY` and `FILTER` clauses. Pypika doesn't support ARRAY_AGG or LIST operations
        by default.
        """
        query_sql = (
            f"LIST({self.from_column.get_sql(**kwargs)} "
            f"ORDER BY {self.order_by_column.get_sql(**kwargs)} {self.direction.value}) "
            f"FILTER (WHERE {self.filter_clause.get_sql(**kwargs)})"
        )
        return format_alias_sql(query_sql, self.alias, **kwargs)


class DuckDBAny(Term):
    def __init__(self, column: Term) -> None:
        super().__init__()
        self.column = column

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns `ANY()`. Pypika doesn't support this function by default.
        """
        query_sql = f"ANY({self.column.get_sql(**kwargs)})"
        return format_alias_sql(query_sql, self.alias, **kwargs)


class DuckDBListFilterNulls(Term):
    def __init__(self, column: Term) -> None:
        super().__init__()
        self.column = column

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns an filtered list with non-null elements. Pypika doesn't support this function by default.
        """
        query_sql = f"array_filter({self.column.get_sql(**kwargs)}, element -> element IS NOT NULL)"
        return format_alias_sql(query_sql, self.alias, **kwargs)


class DuckDBStructPack(Term):
    def __init__(self, *fields: Term) -> None:
        super().__init__()
        self.fields = fields

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns `STRUCT_PACK()`. Pypika doesn't support this function by default.
        """
        query_sql = f"STRUCT_PACK({','.join(field.get_sql(**kwargs) for field in self.fields)})"
        return format_alias_sql(query_sql, self.alias, **kwargs)


class DuckDBListTransform(Term):
    def __init__(self, column: Term, lambda_func: str) -> None:
        super().__init__()
        self.column = column
        self.lambda_func = lambda_func

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns `LIST_TRANSFORM()`. Pypika doesn't support this function by default.
        """
        query_sql = f"LIST_TRANSFORM({self.column.get_sql(**kwargs)}, {self.lambda_func})"
        return format_alias_sql(query_sql, self.alias, **kwargs)


class DuckDBTupleTerm(TupleTerm):
    def get_sql(self, **kwargs):
        return f"row{super().get_sql(**kwargs)}"


class Values(Term):
    def __init__(
        self, values: List[pypika.Tuple], alias: Optional[str] = None, columns: Optional[List[str]] = None
    ) -> None:
        super().__init__()
        self.values = values
        self.alias = alias
        self.columns = columns

    def get_sql(self, **kwargs: Any) -> str:
        """
        Returns a VALUES clause of the format
        VALUES (a, b, c), (d,e,f) Alias(col1, col2, col3)
        """
        sql = "VALUES {values}".format(values=",".join([row.get_sql() for row in self.values]))

        if self.alias is not None:
            columns = ""
            if self.columns is not None:
                columns = "({columns})".format(columns=",".join(self.columns))

            sql = "({values}) {alias}{columns}".format(values=sql, alias=self.alias, columns=columns)

        return sql


class _DuckDB(CompatFunctions):
    @classmethod
    def query(cls) -> typing.Type[Query]:
        return DuckDBQuery

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        return DuckDBStructPack(*[Field(n) for n in field_names])

    @classmethod
    def list(cls, column: str) -> WindowFrameAnalyticFunction:
        return DuckDBList(Field(column))

    @classmethod
    def any(cls, column: str) -> Term:
        return DuckDBAny(Field(column))

    @classmethod
    def list_filter_nulls(cls, column: str) -> Term:
        return DuckDBListFilterNulls(Field(column))

    @classmethod
    def ordered_filtered_list(
        cls, from_column: Term, order_by_column: Term, filter_clause: Term, direction: Order = Order.asc
    ) -> Term:
        return DuckDBOrderedFilteredList(from_column, order_by_column, filter_clause, direction)

    @classmethod
    def list_transform(cls, column: Term, lambda_func: str) -> Term:
        return DuckDBListTransform(column, lambda_func)

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        assert len(field_names) == len(aliases)
        return [getattr(Table(name), field).as_(alias) for (field, alias) in zip(field_names, aliases)]

    @classmethod
    def to_timestamp(cls, time_col: Union[Term, str]) -> Term:
        return Cast(time_col, as_type="timestamptz")

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        return _Spark.date_add(interval, amount, time_field)

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        return Cast(Floor(Function("date_part", "epoch", cls.to_timestamp(timestamp))), as_type="int64")

    @classmethod
    def to_utc(cls, timestamp: Term) -> Term:
        """
        When DuckDB converts a timestamp column to UTC, the timezone is removed if the original column had a
        timezone. But if the original column did not have a timezone, the timezone is set to UTC.
        We need to make sure that timestamp always has a timezone, because in duckdb >= 0.10
        timezone-naive timestamp can't be compared with timezone-aware timestamp.
        Cast to TIMESTAMPTZ relies on the fact that DuckDB session timezone was set to UTC.
        """
        return Cast(Function("timezone", "UTC", timestamp), "timestamptz")

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        """
        Function `to_timestamp` returns timezone-aware results starting version 0.9.0.
        """
        return Function("to_timestamp", unix_timestamp)

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        return Function("_tecton_int_div", a, b)

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        return Function("strftime", timestamp, fmt)


class _Spark(CompatFunctions):
    @classmethod
    def query(cls) -> typing.Type[Query]:
        # Spark (similar to HiveSQL) uses backticks like MySQL
        return MySQLQuery

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        return Function("struct", *[Field(n) for n in field_names])

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        assert len(field_names) == len(aliases)
        return [getattr(Table(name), field).as_(alias) for (field, alias) in zip(field_names, aliases)]

    @classmethod
    def to_timestamp(cls, time_str: str) -> Term:
        return Function("to_timestamp", time_str)

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        if interval == "second":
            return time_field + Interval(seconds=amount)
        elif interval == "millisecond":
            # Note: PyPika has a bug where for microseconds, it does not correctly respect negative values
            if amount >= 0:
                return time_field + Interval(microseconds=amount * 1000)
            else:
                return time_field - Interval(microseconds=abs(amount) * 1000)
        else:
            msg = f"Unexpected date_add interval {interval}"
            raise NotImplementedError(msg)

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        return Function("unix_timestamp", timestamp)

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        return Cast(Function("from_unixtime", unix_timestamp), as_type="timestamp")

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        raise NotImplementedError()

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        raise NotImplementedError()


class _Athena(CompatFunctions):
    @classmethod
    def query(cls) -> typing.Type[Query]:
        # Athena (similar to PrestoSQL) uses doublequotes like Postgres
        return PostgreSQLQuery

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        return Function("row", *[Field(n) for n in field_names])

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        return [getattr(Table(name), f"field{idx}").as_(aliases[idx]) for idx in range(len(aliases))]

    @classmethod
    def to_timestamp(cls, time_str: str) -> Term:
        return Function("from_iso8601_timestamp", time_str)

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        interval = f"'{interval}'"
        return DateAdd(interval, amount, time_field)

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        return Function("to_unixtime", timestamp)

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        return Function("from_unixtime", unix_timestamp)

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        raise NotImplementedError()

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        raise NotImplementedError()


class _Bigquery(CompatFunctions):
    @classmethod
    def query(cls) -> typing.Type[Query]:
        # BigQuery backticks to quote fields, similar to MySQL
        return MySQLQuery

    @classmethod
    def struct(cls, field_names: List[str]) -> Term:
        return Function("STRUCT", *[Field(n) for n in field_names])

    @classmethod
    def struct_extract(
        cls, name: str, field_names: List[str], aliases: List[str], schema: Dict[str, DataType]
    ) -> List[Field]:
        return [Field(f"{name}.{field_names[idx]}").as_(aliases[idx]) for idx in range(len(aliases))]

    @classmethod
    def to_timestamp(cls, time_str: str) -> Term:
        return Function("TIMESTAMP", time_str)

    @classmethod
    def date_add(cls, interval: str, amount: int, time_field: Term) -> Term:
        interval = LiteralValue(f"INTERVAL {amount} {interval}")
        return Function("DATE_ADD", time_field, interval)

    @classmethod
    def to_unixtime(cls, timestamp: Term) -> Term:
        return Function("UNIX_SECONDS", timestamp)

    @classmethod
    def from_unixtime(cls, unix_timestamp: Term) -> Term:
        return Function("TIMESTAMP_SECONDS", unix_timestamp)

    @classmethod
    def strftime(cls, timestamp: Term, fmt: Union[Term, str]) -> Term:
        return Function("FORMAT_TIMESTAMP", fmt, timestamp)

    @classmethod
    def int_div(cls, a: Union[Term, int], b: Union[Term, int]) -> Term:
        return Function("SAFE_DIVIDE", a, b)

    @classmethod
    def ordered_filtered_list(
        cls, from_column: Term, order_by_column: Term, filter_clause: Term, direction: Order = Order.asc
    ) -> Term:
        raise NotImplementedError()

    @classmethod
    def list(cls, column: str) -> WindowFrameAnalyticFunction:
        raise NotImplementedError()

    @classmethod
    def list_transform(cls, column: Term, lambda_func: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def any(cls, column: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def list_filter_nulls(cls, column: str) -> Term:
        raise NotImplementedError()

    @classmethod
    def to_utc(cls, timestamp: Term) -> Term:
        raise NotImplementedError()


class _Torch(CompatFunctions):  # pylint: disable=abstract-method
    pass



import binascii
import os
import re
import zlib
from functools import wraps
from typing import List
from typing import Optional

import pyspark
from google.protobuf.empty_pb2 import Empty

from tecton_core import errors as tecton_core_errors
from tecton_core import function_deserialization
from tecton_core import specs
from tecton_materialization.materialization import _parse_bucket_key_from_uri
from tecton_proto.common import data_type__client_pb2 as data_type_pb2
from tecton_proto.common.schema__client_pb2 import Schema as SchemaProto
from tecton_proto.common.spark_schema__client_pb2 import SparkSchema
from tecton_proto.data.hive_metastore__client_pb2 import ListHiveResult
from tecton_proto.data.remote_spark__client_pb2 import ExecuteRequest
from tecton_proto.data.remote_spark__client_pb2 import ExecuteResult
from tecton_proto.data.remote_spark__client_pb2 import FeatureViewSchemaResponse
from tecton_proto.data.remote_spark__client_pb2 import GetBatchDataSourceFunctionSchema
from tecton_proto.data.remote_spark__client_pb2 import GetFeatureViewSchema
from tecton_proto.data.remote_spark__client_pb2 import GetFileSourceSchema
from tecton_proto.data.remote_spark__client_pb2 import GetHiveTableSchema
from tecton_proto.data.remote_spark__client_pb2 import GetKafkaSourceSchema
from tecton_proto.data.remote_spark__client_pb2 import GetKinesisSourceSchema
from tecton_proto.data.remote_spark__client_pb2 import GetMultipleDataSourceSchemasRequest
from tecton_proto.data.remote_spark__client_pb2 import GetMultipleDataSourceSchemasResponse
from tecton_proto.data.remote_spark__client_pb2 import GetMultipleFeatureViewSchemasRequest
from tecton_proto.data.remote_spark__client_pb2 import GetMultipleFeatureViewSchemasResponse
from tecton_proto.data.remote_spark__client_pb2 import GetQueryPlanInfoForFeatureViewPipeline
from tecton_proto.data.remote_spark__client_pb2 import GetRedshiftTableSchema
from tecton_proto.data.remote_spark__client_pb2 import GetSnowflakeSchema
from tecton_proto.data.remote_spark__client_pb2 import GetStreamDataSourceFunctionSchema
from tecton_proto.data.remote_spark__client_pb2 import GetUnityTableSchema
from tecton_proto.data.remote_spark__client_pb2 import ListHiveDatabases
from tecton_proto.data.remote_spark__client_pb2 import ListHiveTableColumns
from tecton_proto.data.remote_spark__client_pb2 import ListHiveTables
from tecton_proto.data.remote_spark__client_pb2 import QueryPlanInfo
from tecton_proto.data.remote_spark__client_pb2 import SparkSchemas
from tecton_proto.data.virtual_data_source__client_pb2 import VirtualDataSource
from tecton_spark import schema_derivation_utils
from tecton_spark.spark_helper import get_query_plan_info_for_df
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper


# Return raw Spark exceptions as strings to avoid py4j wrapping that obfuscates the root cause
def exception_catching_wrapper(func):
    @wraps(func)
    def _exception_catching_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except pyspark.sql.utils.CapturedException as e:
            msg = e.desc if e.desc else str(e)
            return msg

    return _exception_catching_wrapper


# TODO(rafael): We should move this to a `fs_utils` module at some point and collect any related functionality there.
def _write_to_path(path: str, content: str) -> None:
    if path.startswith("s3://"):
        import boto3  # import here since this is not always available, but should if path is S3

        # NOTE(rafael): This will inherit whatever default access the current role has.
        s3 = boto3.resource("s3")
        bucket_name, key = _parse_bucket_key_from_uri(path)
        s3.Object(bucket_name, key).put(Body=content)
    elif path.startswith("dbfs:/"):
        path = re.sub("^dbfs:/", "/dbfs/", path)
        # create directory if it does not exist
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w") as fid:
            fid.write(content)
    else:
        msg = f"'{path}' is not of supported type"
        raise ValueError(msg)


class PysparkRemoteHost(object):
    def __init__(self, spark_session):
        # A way for users to implement custom logic like our built-in schema_uri functionality for datasource-functions.
        os.environ["_TECTON_SCHEMA_DERIVATION"] = "1"
        self.executions_map = {
            "getHiveTableSchema": lambda r: self._getHiveTableSchema(r.getHiveTableSchema),
            "getUnityTableSchema": lambda r: self._getUnityTableSchema(r.getUnityTableSchema),
            "getRedshiftTableSchema": lambda r: self._getRedshiftTableSchema(r.getRedshiftTableSchema),
            "getSnowflakeSchema": lambda r: self._getSnowflakeSchema(r.getSnowflakeSchema),
            "getFileSourceSchema": lambda r: self._getFileDataSourceSchema(r.getFileSourceSchema),
            "getBatchDataSourceFunctionSchema": lambda r: self._getBatchDataSourceFunctionSchema(
                r.getBatchDataSourceFunctionSchema
            ),
            "getStreamDataSourceFunctionSchema": lambda r: self._getStreamDataSourceFunctionSchema(
                r.getStreamDataSourceFunctionSchema
            ),
            "getKinesisSourceSchema": lambda r: self._getKinesisDataSourceSchema(r.getKinesisSourceSchema),
            "getKafkaSourceSchema": lambda r: self._getKafkaDataSourceSchema(r.getKafkaSourceSchema),
            "getFeatureViewSchema": lambda r: self._getFeatureViewSchema(r.getFeatureViewSchema),
            "getMultipleFeatureViewSchemasRequest": lambda r: self._getMultipleFeatureViewSchemasRequest(
                r.getMultipleFeatureViewSchemasRequest
            ),
            "getMultipleDataSourceSchemasRequest": lambda r: self._getMultipleDataSourceSchemasRequest(
                r.getMultipleDataSourceSchemasRequest
            ),
            "getQueryPlanInfoForFeatureViewPipeline": lambda r: self._getQueryPlanInfoForFeatureViewPipeline(
                r.getQueryPlanInfoForFeatureViewPipeline
            ),
            "listHiveDatabases": lambda r: self._listHiveDatabases(r.listHiveDatabases),
            "listHiveTables": lambda r: self._listHiveTables(r.listHiveTables),
            "listHiveTableColumns": lambda r: self._listHiveTableColumns(r.listHiveTableColumns),
        }

        self.spark = spark_session

    def _getHiveTableSchema(self, request: GetHiveTableSchema) -> SparkSchema:
        if request.HasField("rawBatchTranslator"):
            post_processor = function_deserialization.from_proto_to_main(request.rawBatchTranslator)
        else:
            post_processor = None

        return schema_derivation_utils.get_hive_table_schema(
            spark=self.spark,
            database=request.database,
            table=request.table,
            post_processor=post_processor,
            timestamp_field=request.timestampColumn,
            timestamp_format=request.timestampFormat,
        )

    def _getUnityTableSchema(self, request: GetUnityTableSchema) -> SparkSchema:
        if request.HasField("rawBatchTranslator"):
            post_processor = function_deserialization.from_proto_to_main(request.rawBatchTranslator)
        else:
            post_processor = None

        return schema_derivation_utils.get_unity_table_schema(
            spark=self.spark,
            catalog=request.catalog,
            schema=request.schema,
            table=request.table,
            post_processor=post_processor,
            timestamp_field=request.timestampColumn,
            timestamp_format=request.timestampFormat,
        )

    def _getBatchDataSourceFunctionSchema(self, request: GetBatchDataSourceFunctionSchema) -> SparkSchema:
        data_source_fn = function_deserialization.from_proto_to_main(request.function)
        return schema_derivation_utils.get_batch_data_source_function_schema(
            spark=self.spark,
            data_source_function=data_source_fn,
            supports_time_filtering=request.supports_time_filtering,
        )

    def _getStreamDataSourceFunctionSchema(self, request: GetStreamDataSourceFunctionSchema) -> SparkSchema:
        data_source_fn = function_deserialization.from_proto_to_main(request.function)
        return schema_derivation_utils.get_stream_data_source_function_schema(self.spark, data_source_fn)

    def _getRedshiftTableSchema(self, request: GetRedshiftTableSchema) -> SparkSchema:
        assert request.HasField("endpoint") and request.endpoint, "endpoint cannot be None"
        assert request.HasField("temp_s3_dir") and request.temp_s3_dir, "temp_s3_dir cannot be None"

        assert (request.HasField("table") and request.table) or (
            request.HasField("query") and request.query
        ), "Both table and query cannot be None"
        if request.HasField("rawBatchTranslator"):
            post_processor = function_deserialization.from_proto_to_main(request.rawBatchTranslator)
        else:
            post_processor = None
        return schema_derivation_utils.get_redshift_table_schema(
            spark=self.spark,
            endpoint=request.endpoint,
            table=request.table,
            query=request.query,
            temp_s3=request.temp_s3_dir,
            post_processor=post_processor,
        )

    def _getSnowflakeSchema(self, request: GetSnowflakeSchema) -> SparkSchema:
        assert request.HasField("url"), "url cannot be None"
        assert request.HasField("database"), "database cannot be None"
        assert request.HasField("schema"), "schema cannot be None"
        assert request.HasField("warehouse"), "warehouse cannot be None"

        if request.HasField("post_processor"):
            post_processor = function_deserialization.from_proto_to_main(request.post_processor)
        else:
            post_processor = None

        return schema_derivation_utils.get_snowflake_schema(
            spark=self.spark,
            url=request.url,
            database=request.database,
            schema=request.schema,
            warehouse=request.warehouse,
            role=request.role if request.HasField("role") else None,
            table=request.table if request.HasField("table") else None,
            query=request.query if request.HasField("query") else None,
            post_processor=post_processor,
        )

    def _getFileDataSourceSchema(self, request: GetFileSourceSchema) -> SparkSchema:
        schema_override = None
        if request.HasField("schemaOverride"):
            schema_override = SparkSchemaWrapper.from_proto(request.schemaOverride)

        post_processor = None
        if request.HasField("rawBatchTranslator"):
            post_processor = function_deserialization.from_proto_to_main(request.rawBatchTranslator)

        return schema_derivation_utils.get_file_source_schema(
            spark=self.spark,
            file_format=request.fileFormat,
            file_uri=request.uri,
            convert_to_glue=request.convertToGlueFormat,
            schema_uri=request.schemaUri if request.HasField("schemaUri") else None,
            schema_override=schema_override,
            post_processor=post_processor,
            timestamp_col=request.timestampColumn if request.HasField("timestampColumn") else None,
            timestmap_format=request.timestampFormat if request.HasField("timestampFormat") else None,
        )

    def _getKinesisDataSourceSchema(self, request: GetKinesisSourceSchema) -> SparkSchema:
        translator_fn = function_deserialization.from_proto_to_main(request.rawStreamTranslator)
        return schema_derivation_utils.get_kinesis_schema(self.spark, request.streamName, translator_fn)

    def _getKafkaDataSourceSchema(self, request: GetKafkaSourceSchema) -> SparkSchema:
        translator_fn = function_deserialization.from_proto_to_main(request.rawStreamTranslator)
        return schema_derivation_utils.get_kafka_schema(self.spark, translator_fn)

    @classmethod
    def deserialize_virtual_data_sources(cls, virtual_data_sources_bin: List[bytes]) -> List[VirtualDataSource]:
        data_sources = []
        for bstr in virtual_data_sources_bin:
            data_source = VirtualDataSource()
            data_source.ParseFromString(bstr)
            data_sources.append(data_source)
        return data_sources

    def _getMultipleDataSourceSchemasRequest(
        self, request: GetMultipleDataSourceSchemasRequest
    ) -> GetMultipleDataSourceSchemasResponse:
        batched_responses = GetMultipleDataSourceSchemasResponse()
        for ds in request.data_source_args:
            response = SparkSchemas()
            # Create a spec to leverage its unified interface. (Usually specs are only created from fully validated args with derived schemas.)
            spec = specs.DataSourceSpec.from_args_proto(
                ds,
                specs.DataSourceSpecArgsSupplement(
                    # The spec constructor expects a non-null schema
                    batch_schema=SparkSchema(),
                    stream_schema=SparkSchema(),
                ),
            )
            if spec.batch_source is not None:
                response.batchSchema.CopyFrom(
                    schema_derivation_utils.derive_batch_schema(
                        self.spark,
                        ds,
                        getattr(spec.batch_source, "post_processor", None),
                        getattr(spec.batch_source, "function", None),
                    )
                )
            if spec.stream_source is not None:
                response.streamSchema.CopyFrom(
                    schema_derivation_utils.derive_stream_schema(
                        self.spark,
                        ds,
                        getattr(spec.stream_source, "post_processor", None),
                        getattr(spec.stream_source, "function", None),
                    )
                )
            batched_responses.data_source_responses[ds.info.name].CopyFrom(response)
        return batched_responses

    def _getMultipleFeatureViewSchemasRequest(
        self, request: GetMultipleFeatureViewSchemasRequest
    ) -> GetMultipleFeatureViewSchemasResponse:
        data_source_specs = [
            specs.DataSourceSpec.from_data_proto(ds, deserialize_funcs_to_main=True)
            for ds in request.virtual_data_sources
        ]
        transformation_specs = [
            specs.TransformationSpec.from_data_proto(t, deserialize_funcs_to_main=True) for t in request.transformations
        ]
        batched_responses = GetMultipleFeatureViewSchemasResponse()
        for fv_request in request.feature_view_requests:
            response = FeatureViewSchemaResponse()
            response.view_schema.CopyFrom(
                schema_derivation_utils.get_feature_view_view_schema(
                    self.spark,
                    fv_request.feature_view,
                    transformation_specs,
                    data_source_specs,
                )
            )
            timestamp_keys = [
                c.name
                for c in response.view_schema.columns
                if c.offline_data_type.type == data_type_pb2.DATA_TYPE_TIMESTAMP
            ]
            if len(timestamp_keys) != 1:
                msg = f"Expected single timestamp column for Feature View {fv_request.feature_view.info.name}. Found {timestamp_keys}"
                raise tecton_core_errors.TectonValidationError(msg)
            else:
                response.timestamp_key = timestamp_keys[0]
            batched_responses.feature_view_responses[fv_request.feature_view.info.name].CopyFrom(response)
        return batched_responses

    def _getFeatureViewSchema(self, request: GetFeatureViewSchema) -> SchemaProto:
        data_source_specs = [
            specs.DataSourceSpec.from_data_proto(ds, deserialize_funcs_to_main=True)
            for ds in request.virtual_data_sources
        ]
        transformation_specs = [
            specs.TransformationSpec.from_data_proto(t, deserialize_funcs_to_main=True) for t in request.transformations
        ]
        return schema_derivation_utils.get_feature_view_view_schema(
            self.spark, request.feature_view, transformation_specs, data_source_specs
        )

    def _getQueryPlanInfoForFeatureViewPipeline(self, request: GetQueryPlanInfoForFeatureViewPipeline) -> QueryPlanInfo:
        data_source_specs = [
            specs.DataSourceSpec.from_data_proto(ds, deserialize_funcs_to_main=True)
            for ds in request.virtual_data_sources
        ]
        transformation_specs = [
            specs.TransformationSpec.from_data_proto(t, deserialize_funcs_to_main=True) for t in request.transformations
        ]
        df = schema_derivation_utils.get_feature_view_empty_view_df(
            self.spark, request.feature_view, transformation_specs, data_source_specs
        )

        query_plan = get_query_plan_info_for_df(df)

        return QueryPlanInfo(
            has_joins=query_plan.has_joins,
            has_aggregations=query_plan.has_aggregations,
        )

    def _listHiveDatabases(self, request: ListHiveDatabases) -> ListHiveResult:
        spark_sql = self.spark.sql("SHOW DATABASES")
        return ListHiveResult(names=[row["databaseName"] for row in spark_sql.collect()])

    def _listHiveTables(self, request: ListHiveTables) -> ListHiveResult:
        spark_sql = self.spark.sql(f"SHOW TABLES FROM `{request.database}`")
        return ListHiveResult(names=[row["tableName"] for row in spark_sql.collect()])

    def _listHiveTableColumns(self, request: ListHiveTableColumns) -> ListHiveResult:
        spark_sql = self.spark.sql(f"DESCRIBE {request.database}.{request.table}")
        return ListHiveResult(names=[row["col_name"] for row in spark_sql.collect()])

    def execute(self, request: str, output_path: Optional[str] = None) -> Optional[str]:
        request_bytes = binascii.unhexlify(request)
        execute_request = ExecuteRequest()
        execute_request.ParseFromString(request_bytes)

        request_type = execute_request.WhichOneof("request")
        if request_type is None:
            execute_result = ExecuteResult()
            execute_result.errorMessage = "Missing request in ExecuteRequest message"
            return str(binascii.b2a_base64(execute_result.SerializeToString()).rstrip())

        try:
            for k, v in execute_request.envVars.items():
                os.environ[k] = v
            handler = self.executions_map[request_type]
            resp = handler(execute_request)
            result = self._serialize_response(resp)
        except tecton_core_errors.TectonValidationError as e:
            result = ExecuteResult()
            result.validationError = str(e)
        except pyspark.sql.utils.CapturedException as e:
            import traceback

            result = ExecuteResult()
            result.uncaughtError = (
                e.desc if e.desc else "".join(traceback.TracebackException.from_exception(e).format())
            )
        except Exception as e:
            import traceback

            result = ExecuteResult()
            result.uncaughtError = "".join(traceback.TracebackException.from_exception(e).format())

        resp = str(binascii.b2a_base64(zlib.compress(result.SerializeToString())).rstrip())
        if output_path is None:
            return resp
        else:
            _write_to_path(path=output_path, content=resp)

    @staticmethod
    def _serialize_response(result) -> ExecuteResult:
        if isinstance(result, Empty):
            return ExecuteResult()
        elif isinstance(result, SparkSchema):
            return ExecuteResult(sparkSchema=result)
        elif isinstance(result, QueryPlanInfo):
            return ExecuteResult(queryPlanInfo=result)
        elif isinstance(result, SchemaProto):
            return ExecuteResult(schema=result)
        elif isinstance(result, ListHiveResult):
            return ExecuteResult(listHiveResult=result)
        elif isinstance(result, GetMultipleFeatureViewSchemasResponse):
            return ExecuteResult(multipleFeatureViewSchemaResponse=result)
        elif isinstance(result, GetMultipleDataSourceSchemasResponse):
            return ExecuteResult(multipleDataSourceSchemaResponse=result)
        msg = f"Unidentified return type: ${type(result)}"
        raise Exception(msg)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

from pyspark.sql.column import Column, _to_seq
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.pandas.group_ops import PandasGroupedOpsMixin
from pyspark.sql.types import StructType, StructField, IntegerType, StringType

__all__ = ["GroupedData"]


def dfapi(f):
    def _api(self):
        name = f.__name__
        jdf = getattr(self._jgd, name)()
        return DataFrame(jdf, self.sql_ctx)
    _api.__name__ = f.__name__
    _api.__doc__ = f.__doc__
    return _api


def df_varargs_api(f):
    def _api(self, *cols):
        name = f.__name__
        jdf = getattr(self._jgd, name)(_to_seq(self.sql_ctx._sc, cols))
        return DataFrame(jdf, self.sql_ctx)
    _api.__name__ = f.__name__
    _api.__doc__ = f.__doc__
    return _api


class GroupedData(PandasGroupedOpsMixin):
    """
    A set of methods for aggregations on a :class:`DataFrame`,
    created by :func:`DataFrame.groupBy`.

    .. versionadded:: 1.3
    """

    def __init__(self, jgd, df):
        self._jgd = jgd
        self._df = df
        self.sql_ctx = df.sql_ctx

    def agg(self, *exprs):
        """Compute aggregates and returns the result as a :class:`DataFrame`.

        The available aggregate functions can be:

        1. built-in aggregation functions, such as `avg`, `max`, `min`, `sum`, `count`

        2. group aggregate pandas UDFs, created with :func:`pyspark.sql.functions.pandas_udf`

           .. note:: There is no partial aggregation with group aggregate UDFs, i.e.,
               a full shuffle is required. Also, all the data of a group will be loaded into
               memory, so the user should be aware of the potential OOM risk if data is skewed
               and certain groups are too large to fit in memory.

           .. seealso:: :func:`pyspark.sql.functions.pandas_udf`

        If ``exprs`` is a single :class:`dict` mapping from string to string, then the key
        is the column to perform aggregation on, and the value is the aggregate function.

        Alternatively, ``exprs`` can also be a list of aggregate :class:`Column` expressions.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        exprs : dict
            a dict mapping from column name (string) to aggregate functions (string),
            or a list of :class:`Column`.

        Notes
        -----
        Built-in aggregation functions and group aggregate pandas UDFs cannot be mixed
        in a single call to this function.

        Examples
        --------
        >>> gdf = df.groupBy(df.name)
        >>> sorted(gdf.agg({"*": "count"}).collect())
        [Row(name='Alice', count(1)=1), Row(name='Bob', count(1)=1)]

        >>> from pyspark.sql import functions as F
        >>> sorted(gdf.agg(F.min(df.age)).collect())
        [Row(name='Alice', min(age)=2), Row(name='Bob', min(age)=5)]

        >>> from pyspark.sql.functions import pandas_udf, PandasUDFType
        >>> @pandas_udf('int', PandasUDFType.GROUPED_AGG)  # doctest: +SKIP
        ... def min_udf(v):
        ...     return v.min()
        >>> sorted(gdf.agg(min_udf(df.age)).collect())  # doctest: +SKIP
        [Row(name='Alice', min_udf(age)=2), Row(name='Bob', min_udf(age)=5)]
        """
        assert exprs, "exprs should not be empty"
        if len(exprs) == 1 and isinstance(exprs[0], dict):
            jdf = self._jgd.agg(exprs[0])
        else:
            # Columns
            assert all(isinstance(c, Column) for c in exprs), "all exprs should be Column"
            jdf = self._jgd.agg(exprs[0]._jc,
                                _to_seq(self.sql_ctx._sc, [c._jc for c in exprs[1:]]))
        return DataFrame(jdf, self.sql_ctx)

    @dfapi
    def count(self):
        """Counts the number of records for each group.

        .. versionadded:: 1.3.0

        Examples
        --------
        >>> sorted(df.groupBy(df.age).count().collect())
        [Row(age=2, count=1), Row(age=5, count=1)]
        """

    @df_varargs_api
    def mean(self, *cols):
        """Computes average values for each numeric columns for each group.

        :func:`mean` is an alias for :func:`avg`.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples
        --------
        >>> df.groupBy().mean('age').collect()
        [Row(avg(age)=3.5)]
        >>> df3.groupBy().mean('age', 'height').collect()
        [Row(avg(age)=3.5, avg(height)=82.5)]
        """

    @df_varargs_api
    def avg(self, *cols):
        """Computes average values for each numeric columns for each group.

        :func:`mean` is an alias for :func:`avg`.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples
        --------
        >>> df.groupBy().avg('age').collect()
        [Row(avg(age)=3.5)]
        >>> df3.groupBy().avg('age', 'height').collect()
        [Row(avg(age)=3.5, avg(height)=82.5)]
        """

    @df_varargs_api
    def max(self, *cols):
        """Computes the max value for each numeric columns for each group.

        .. versionadded:: 1.3.0

        Examples
        --------
        >>> df.groupBy().max('age').collect()
        [Row(max(age)=5)]
        >>> df3.groupBy().max('age', 'height').collect()
        [Row(max(age)=5, max(height)=85)]
        """

    @df_varargs_api
    def min(self, *cols):
        """Computes the min value for each numeric column for each group.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples
        --------
        >>> df.groupBy().min('age').collect()
        [Row(min(age)=2)]
        >>> df3.groupBy().min('age', 'height').collect()
        [Row(min(age)=2, min(height)=80)]
        """

    @df_varargs_api
    def sum(self, *cols):
        """Compute the sum for each numeric columns for each group.

        .. versionadded:: 1.3.0

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples
        --------
        >>> df.groupBy().sum('age').collect()
        [Row(sum(age)=7)]
        >>> df3.groupBy().sum('age', 'height').collect()
        [Row(sum(age)=7, sum(height)=165)]
        """

    def pivot(self, pivot_col, values=None):
        """
        Pivots a column of the current :class:`DataFrame` and perform the specified aggregation.
        There are two versions of pivot function: one that requires the caller to specify the list
        of distinct values to pivot on, and one that does not. The latter is more concise but less
        efficient, because Spark needs to first compute the list of distinct values internally.

        .. versionadded:: 1.6.0

        Parameters
        ----------
        pivot_col : str
            Name of the column to pivot.
        values :
            List of values that will be translated to columns in the output DataFrame.

        Examples
        --------
        # Compute the sum of earnings for each year by course with each course as a separate column

        >>> df4.groupBy("year").pivot("course", ["dotNET", "Java"]).sum("earnings").collect()
        [Row(year=2012, dotNET=15000, Java=20000), Row(year=2013, dotNET=48000, Java=30000)]

        # Or without specifying column values (less efficient)

        >>> df4.groupBy("year").pivot("course").sum("earnings").collect()
        [Row(year=2012, Java=20000, dotNET=15000), Row(year=2013, Java=30000, dotNET=48000)]
        >>> df5.groupBy("sales.year").pivot("sales.course").sum("sales.earnings").collect()
        [Row(year=2012, Java=20000, dotNET=15000), Row(year=2013, Java=30000, dotNET=48000)]
        """
        if values is None:
            jgd = self._jgd.pivot(pivot_col)
        else:
            jgd = self._jgd.pivot(pivot_col, values)
        return GroupedData(jgd, self._df)


def _test():
    import doctest
    from pyspark.sql import Row, SparkSession
    import pyspark.sql.group
    globs = pyspark.sql.group.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("sql.group tests")\
        .getOrCreate()
    sc = spark.sparkContext
    globs['sc'] = sc
    globs['spark'] = spark
    globs['df'] = sc.parallelize([(2, 'Alice'), (5, 'Bob')]) \
        .toDF(StructType([StructField('age', IntegerType()),
                          StructField('name', StringType())]))
    globs['df3'] = sc.parallelize([Row(name='Alice', age=2, height=80),
                                   Row(name='Bob', age=5, height=85)]).toDF()
    globs['df4'] = sc.parallelize([Row(course="dotNET", year=2012, earnings=10000),
                                   Row(course="Java",   year=2012, earnings=20000),
                                   Row(course="dotNET", year=2012, earnings=5000),
                                   Row(course="dotNET", year=2013, earnings=48000),
                                   Row(course="Java",   year=2013, earnings=30000)]).toDF()
    globs['df5'] = sc.parallelize([
        Row(training="expert", sales=Row(course="dotNET", year=2012, earnings=10000)),
        Row(training="junior", sales=Row(course="Java",   year=2012, earnings=20000)),
        Row(training="expert", sales=Row(course="dotNET", year=2012, earnings=5000)),
        Row(training="junior", sales=Row(course="dotNET", year=2013, earnings=48000)),
        Row(training="expert", sales=Row(course="Java",   year=2013, earnings=30000))]).toDF()

    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.group, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import gc
import os
import sys
from tempfile import NamedTemporaryFile
import threading
import pickle

from pyspark.java_gateway import local_connect_and_auth
from pyspark.serializers import ChunkedStream, pickle_protocol
from pyspark.util import print_exec


__all__ = ['Broadcast']


# Holds broadcasted data received from Java, keyed by its id.
_broadcastRegistry = {}


def _from_id(bid):
    from pyspark.broadcast import _broadcastRegistry
    if bid not in _broadcastRegistry:
        raise Exception("Broadcast variable '%s' not loaded!" % bid)
    return _broadcastRegistry[bid]


class Broadcast(object):

    """
    A broadcast variable created with :meth:`SparkContext.broadcast`.
    Access its value through :attr:`value`.

    Examples
    --------
    >>> from pyspark.context import SparkContext
    >>> sc = SparkContext('local', 'test')
    >>> b = sc.broadcast([1, 2, 3, 4, 5])
    >>> b.value
    [1, 2, 3, 4, 5]
    >>> sc.parallelize([0, 0]).flatMap(lambda x: b.value).collect()
    [1, 2, 3, 4, 5, 1, 2, 3, 4, 5]
    >>> b.unpersist()

    >>> large_broadcast = sc.broadcast(range(10000))
    """

    def __init__(self, sc=None, value=None, pickle_registry=None, path=None,
                 sock_file=None):
        """
        Should not be called directly by users -- use :meth:`SparkContext.broadcast`
        instead.
        """
        if sc is not None:
            # we're on the driver.  We want the pickled data to end up in a file (maybe encrypted)
            f = NamedTemporaryFile(delete=False, dir=sc._temp_dir)
            self._path = f.name
            self._sc = sc
            self._python_broadcast = sc._jvm.PythonRDD.setupBroadcast(self._path)
            if sc._encryption_enabled:
                # with encryption, we ask the jvm to do the encryption for us, we send it data
                # over a socket
                port, auth_secret = self._python_broadcast.setupEncryptionServer()
                (encryption_sock_file, _) = local_connect_and_auth(port, auth_secret)
                broadcast_out = ChunkedStream(encryption_sock_file, 8192)
            else:
                # no encryption, we can just write pickled data directly to the file from python
                broadcast_out = f
            self.dump(value, broadcast_out)
            if sc._encryption_enabled:
                self._python_broadcast.waitTillDataReceived()
            self._jbroadcast = sc._jsc.broadcast(self._python_broadcast)
            self._pickle_registry = pickle_registry
        else:
            # we're on an executor
            self._jbroadcast = None
            self._sc = None
            self._python_broadcast = None
            if sock_file is not None:
                # the jvm is doing decryption for us.  Read the value
                # immediately from the sock_file
                self._value = self.load(sock_file)
            else:
                # the jvm just dumps the pickled data in path -- we'll unpickle lazily when
                # the value is requested
                assert(path is not None)
                self._path = path

    def dump(self, value, f):
        try:
            pickle.dump(value, f, pickle_protocol)
        except pickle.PickleError:
            raise
        except Exception as e:
            msg = "Could not serialize broadcast: %s: %s" \
                  % (e.__class__.__name__, str(e))
            print_exec(sys.stderr)
            raise pickle.PicklingError(msg)
        f.close()

    def load_from_path(self, path):
        with open(path, 'rb', 1 << 20) as f:
            return self.load(f)

    def load(self, file):
        # "file" could also be a socket
        gc.disable()
        try:
            return pickle.load(file)
        finally:
            gc.enable()

    @property
    def value(self):
        """ Return the broadcasted value
        """
        if not hasattr(self, "_value") and self._path is not None:
            # we only need to decrypt it here when encryption is enabled and
            # if its on the driver, since executor decryption is handled already
            if self._sc is not None and self._sc._encryption_enabled:
                port, auth_secret = self._python_broadcast.setupDecryptionServer()
                (decrypted_sock_file, _) = local_connect_and_auth(port, auth_secret)
                self._python_broadcast.waitTillBroadcastDataSent()
                return self.load(decrypted_sock_file)
            else:
                self._value = self.load_from_path(self._path)
        return self._value

    def unpersist(self, blocking=False):
        """
        Delete cached copies of this broadcast on the executors. If the
        broadcast is used after this is called, it will need to be
        re-sent to each executor.

        Parameters
        ----------
        blocking : bool, optional
            Whether to block until unpersisting has completed
        """
        if self._jbroadcast is None:
            raise Exception("Broadcast can only be unpersisted in driver")
        self._jbroadcast.unpersist(blocking)

    def destroy(self, blocking=False):
        """
        Destroy all data and metadata related to this broadcast variable.
        Use this with caution; once a broadcast variable has been destroyed,
        it cannot be used again.

        .. versionchanged:: 3.0.0
           Added optional argument `blocking` to specify whether to block until all
           blocks are deleted.

        Parameters
        ----------
        blocking : bool, optional
            Whether to block until unpersisting has completed
        """
        if self._jbroadcast is None:
            raise Exception("Broadcast can only be destroyed in driver")
        self._jbroadcast.destroy(blocking)
        os.unlink(self._path)

    def __reduce__(self):
        if self._jbroadcast is None:
            raise Exception("Broadcast can only be serialized in driver")
        self._pickle_registry.add(self)
        return _from_id, (self._jbroadcast.id(),)


class BroadcastPickleRegistry(threading.local):
    """ Thread-local registry for broadcast variables that have been pickled
    """

    def __init__(self):
        self.__dict__.setdefault("_registry", set())

    def __iter__(self):
        for bcast in self._registry:
            yield bcast

    def add(self, bcast):
        self._registry.add(bcast)

    def clear(self):
        self._registry.clear()


if __name__ == "__main__":
    import doctest
    (failure_count, test_count) = doctest.testmod()
    if failure_count:
        sys.exit(-1)



from copy import deepcopy
from datetime import datetime
from datetime import timezone
from typing import Dict
from typing import List
from typing import Tuple

from google.protobuf import timestamp_pb2
from jinja2 import BaseLoader
from jinja2 import Environment
from texttable import Texttable

from tecton_proto.data.summary__client_pb2 import FcoSummary
from tecton_proto.data.summary__client_pb2 import SummaryItem


_TABLE_HTML_TEMPLATE = """
<table>
<thead>
    <tr>
    {% for heading in headings %}
    <th style='text-align: left;'>{{heading}}</th>
    {% endfor %}
    </tr>
</thead>
<tbody>
    {% for item in items %}
    <tr>
        {% for value in item %}
        {% if value is string and '\n' in value %}
        <td style='text-align: left;'><pre style='padding: 5px; border:1px solid #efefef'>{{value}}</pre></td>
        {% else %}
        <td style='text-align: left;'>{{value}}</td>
        {% endif %}
        {% endfor %}
    </tr>
    {% endfor %}
</tbody>
</table>
"""


def _fco_metadata_formatter(key, value):
    if isinstance(value, timestamp_pb2.Timestamp):
        t = datetime.fromtimestamp(value.ToSeconds())
        return t.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
    if key == "featureStartTimeSeconds":
        t = datetime.fromtimestamp(int(value))
        return t.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
    if key == "tags":
        return dict(value)
    return value


class Displayable(object):
    """Class that can display data in string, HTML, or dictionary format."""

    _table_template = Environment(loader=BaseLoader(), autoescape=True).from_string(_TABLE_HTML_TEMPLATE)

    def __init__(self, text_table: Texttable, html: str, data: Dict):
        self._text_table = text_table
        self._html = html
        self._data = data

    def __repr__(self):
        return self._text_table.draw()

    def __str__(self):
        return self.__repr__()

    def _repr_html_(self):
        return self._html

    def to_dict(self) -> Dict:
        return deepcopy(self._data)

    @classmethod
    def from_table(
        cls, headings: List[str], rows: List[Tuple], max_width=None, center_align: bool = False
    ) -> "Displayable":
        """
        Returns a Displayable based on tabular data containing an arbitrary number of rows.

        :param headings: N-element list of heading names.
        :param rows: List of N-element tuples containing each row's data.
        :max_width: Optional max-width for the table. Setting to zero creates an unlimited width table.
        """
        for row_index, row in enumerate(rows):
            if len(row) != len(headings):
                msg = f"malformed dimensions: headings has length {len(headings)}, but row {row_index} has length {len(row)}"
                raise ValueError(msg)

        text_table = Texttable()
        if max_width is not None:
            text_table.set_max_width(max_width)
        text_table.add_rows(rows, header=False)
        text_table.header(headings)
        text_table.set_deco(Texttable.HEADER)
        if center_align:
            # Align columns in the middle horizontally
            text_table.set_cols_align(["c" for _ in range(len(headings))])

        # Build HTML version.
        html = cls._table_template.render(headings=headings, items=rows)

        # Build dict version.
        data = {"data": []}
        for row in rows:
            data["data"].append({headings[i]: row[i] for i in range(len(headings))})

        return cls(text_table, html, data)

    @classmethod
    def from_properties(cls, items: List[Tuple]) -> "Displayable":
        """
        Returns a Displayable based on key-value properties.

        :param items: List of 2-element tuples.
        """
        for row_index, item in enumerate(items):
            if len(item) != 2:
                msg = f"row {row_index} has length {len(item)}, but should be 2"
                raise ValueError(msg)

        text_table = Texttable()
        text_table.add_rows(items, header=False)
        text_table.set_deco(Texttable.BORDER | Texttable.VLINES | Texttable.HLINES)

        # build HTML version
        html = cls._table_template.render(items=items)

        # Build dictionary version.
        # E.g. { "k1": "v1", "k2", "v2"}
        data = {item[0]: item[1] for item in items}
        return cls(text_table, html, data)

    @classmethod
    def from_fco_summary(
        cls,
        fco_summary: FcoSummary,
        additional_items: List[Tuple] = [],
    ) -> "Displayable":
        """
        Returns a Displayable based on an FcoSummary proto.

        :param fco_summary: FcoSummary proto.
        :param additional_items: Optional additional 2-element tuples to be displayed.
        """
        data = cls._fco_summary_to_dict(fco_summary)
        for key, value in additional_items:
            data[key] = _fco_metadata_formatter(key, value)

        table_items = []
        for display_name, v in data.items():
            value = None
            if isinstance(v, list):
                value = ", ".join(v)
                table_items.append((display_name, value))
            elif isinstance(v, dict):
                # Render nested tables
                value = cls._draw_nested_table_from_dict(v)
                table_items.append((display_name, value))
            elif isinstance(v, str):
                table_items.append((display_name, v))
            else:
                msg = f"Unsupported type {type(v)}"
                raise TypeError(msg)

        text_table = Texttable()
        text_table.add_rows(table_items, header=False)
        text_table.set_deco(Texttable.BORDER | Texttable.VLINES | Texttable.HLINES)
        text_table.set_cols_valign(["m", "m"])

        # build HTML version
        html = cls._table_template.render(items=table_items)

        return cls(text_table, html, data)

    @classmethod
    def _fco_summary_to_dict(cls, fco_summary: FcoSummary) -> Dict:
        fco_metadata_fields = [
            "name",
            "workspace",
            "description",
            "created_at",
            "owner",
            "last_modified_by",
            "source_filename",
            "tags",
        ]
        metadata = {
            _transform_field_name(field): _fco_metadata_formatter(field, getattr(fco_summary.fco_metadata, field))
            for field in fco_metadata_fields
        }
        summarydata = cls._get_nested_dict(list(fco_summary.summary_items))
        return {**metadata, **summarydata}

    @classmethod
    def _get_nested_dict(cls, nested_summary_items: List[SummaryItem]) -> Dict:
        data = {}
        for item in nested_summary_items:
            if not item.HasField("display_name"):
                continue
            if item.HasField("value"):
                data_value = _fco_metadata_formatter(item.key, item.value)
            elif len(item.multi_values) > 0:
                data_value = list(item.multi_values)
            elif len(item.nested_summary_items) > 0:
                data_value = cls._get_nested_dict(list(item.nested_summary_items))
            else:
                data_value = ""
            data[item.display_name] = data_value
        return data

    @classmethod
    def _draw_nested_table_from_dict(cls, data: Dict):
        table_items = []
        for display_name, v in data.items():
            value = None
            if isinstance(v, list):
                value = ", ".join(v)
                table_items.append((display_name, value))
            elif isinstance(v, dict):
                value = str(cls._draw_nested_table_from_dict(v))
                table_items.append((display_name, value))
            elif isinstance(v, str):
                table_items.append((display_name, v))
            else:
                msg = f"Unsupported type {type(v)}"
                raise TypeError(msg)

        text_table = Texttable()
        text_table.add_rows(table_items, header=False)
        text_table.set_deco(Texttable.HLINES)
        return text_table.draw()


def _transform_field_name(name):
    return name.replace("_", " ").title()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Type-specific codes between pandas and PyArrow. Also contains some utils to correct
pandas instances during the type conversion.
"""

from pyspark.sql.types import BooleanType, ByteType, ShortType, IntegerType, LongType, \
    FloatType, DoubleType, DecimalType, StringType, BinaryType, DateType, TimestampType, \
    ArrayType, MapType, StructType, StructField


def to_arrow_type(dt):
    """ Convert Spark data type to pyarrow type
    """
    from distutils.version import LooseVersion
    import pyarrow as pa
    if type(dt) == BooleanType:
        arrow_type = pa.bool_()
    elif type(dt) == ByteType:
        arrow_type = pa.int8()
    elif type(dt) == ShortType:
        arrow_type = pa.int16()
    elif type(dt) == IntegerType:
        arrow_type = pa.int32()
    elif type(dt) == LongType:
        arrow_type = pa.int64()
    elif type(dt) == FloatType:
        arrow_type = pa.float32()
    elif type(dt) == DoubleType:
        arrow_type = pa.float64()
    elif type(dt) == DecimalType:
        arrow_type = pa.decimal128(dt.precision, dt.scale)
    elif type(dt) == StringType:
        arrow_type = pa.string()
    elif type(dt) == BinaryType:
        arrow_type = pa.binary()
    elif type(dt) == DateType:
        arrow_type = pa.date32()
    elif type(dt) == TimestampType:
        # Timestamps should be in UTC, JVM Arrow timestamps require a timezone to be read
        arrow_type = pa.timestamp('us', tz='UTC')
    elif type(dt) == ArrayType:
        if type(dt.elementType) in [StructType, TimestampType]:
            raise TypeError("Unsupported type in conversion to Arrow: " + str(dt))
        arrow_type = pa.list_(to_arrow_type(dt.elementType))
    elif type(dt) == MapType:
        if LooseVersion(pa.__version__) < LooseVersion("2.0.0"):
            raise TypeError("MapType is only supported with pyarrow 2.0.0 and above")
        if type(dt.keyType) in [StructType, TimestampType] or \
                type(dt.valueType) in [StructType, TimestampType]:
            raise TypeError("Unsupported type in conversion to Arrow: " + str(dt))
        arrow_type = pa.map_(to_arrow_type(dt.keyType), to_arrow_type(dt.valueType))
    elif type(dt) == StructType:
        if any(type(field.dataType) == StructType for field in dt):
            raise TypeError("Nested StructType not supported in conversion to Arrow")
        fields = [pa.field(field.name, to_arrow_type(field.dataType), nullable=field.nullable)
                  for field in dt]
        arrow_type = pa.struct(fields)
    else:
        raise TypeError("Unsupported type in conversion to Arrow: " + str(dt))
    return arrow_type


def to_arrow_schema(schema):
    """ Convert a schema from Spark to Arrow
    """
    import pyarrow as pa
    fields = [pa.field(field.name, to_arrow_type(field.dataType), nullable=field.nullable)
              for field in schema]
    return pa.schema(fields)


def from_arrow_type(at):
    """ Convert pyarrow type to Spark data type.
    """
    from distutils.version import LooseVersion
    import pyarrow as pa
    import pyarrow.types as types
    if types.is_boolean(at):
        spark_type = BooleanType()
    elif types.is_int8(at):
        spark_type = ByteType()
    elif types.is_int16(at):
        spark_type = ShortType()
    elif types.is_int32(at):
        spark_type = IntegerType()
    elif types.is_int64(at):
        spark_type = LongType()
    elif types.is_float32(at):
        spark_type = FloatType()
    elif types.is_float64(at):
        spark_type = DoubleType()
    elif types.is_decimal(at):
        spark_type = DecimalType(precision=at.precision, scale=at.scale)
    elif types.is_string(at):
        spark_type = StringType()
    elif types.is_binary(at):
        spark_type = BinaryType()
    elif types.is_date32(at):
        spark_type = DateType()
    elif types.is_timestamp(at):
        spark_type = TimestampType()
    elif types.is_list(at):
        if types.is_timestamp(at.value_type):
            raise TypeError("Unsupported type in conversion from Arrow: " + str(at))
        spark_type = ArrayType(from_arrow_type(at.value_type))
    elif types.is_map(at):
        if LooseVersion(pa.__version__) < LooseVersion("2.0.0"):
            raise TypeError("MapType is only supported with pyarrow 2.0.0 and above")
        if types.is_timestamp(at.key_type) or types.is_timestamp(at.item_type):
            raise TypeError("Unsupported type in conversion from Arrow: " + str(at))
        spark_type = MapType(from_arrow_type(at.key_type), from_arrow_type(at.item_type))
    elif types.is_struct(at):
        if any(types.is_struct(field.type) for field in at):
            raise TypeError("Nested StructType not supported in conversion from Arrow: " + str(at))
        return StructType(
            [StructField(field.name, from_arrow_type(field.type), nullable=field.nullable)
             for field in at])
    elif types.is_dictionary(at):
        spark_type = from_arrow_type(at.value_type)
    else:
        raise TypeError("Unsupported type in conversion from Arrow: " + str(at))
    return spark_type


def from_arrow_schema(arrow_schema):
    """ Convert schema from Arrow to Spark.
    """
    return StructType(
        [StructField(field.name, from_arrow_type(field.type), nullable=field.nullable)
         for field in arrow_schema])


def _get_local_timezone():
    """ Get local timezone using pytz with environment variable, or dateutil.

    If there is a 'TZ' environment variable, pass it to pandas to use pytz and use it as timezone
    string, otherwise use the special word 'dateutil/:' which means that pandas uses dateutil and
    it reads system configuration to know the system local timezone.

    See also:
    - https://github.com/pandas-dev/pandas/blob/0.19.x/pandas/tslib.pyx#L1753
    - https://github.com/dateutil/dateutil/blob/2.6.1/dateutil/tz/tz.py#L1338
    """
    import os
    return os.environ.get('TZ', 'dateutil/:')


def _check_series_localize_timestamps(s, timezone):
    """
    Convert timezone aware timestamps to timezone-naive in the specified timezone or local timezone.

    If the input series is not a timestamp series, then the same series is returned. If the input
    series is a timestamp series, then a converted series is returned.

    Parameters
    ----------
    s : pandas.Series
    timezone : str
        the timezone to convert. if None then use local timezone

    Returns
    -------
    pandas.Series
        `pandas.Series` that have been converted to tz-naive
    """
    from pyspark.sql.pandas.utils import require_minimum_pandas_version
    require_minimum_pandas_version()

    from pandas.api.types import is_datetime64tz_dtype
    tz = timezone or _get_local_timezone()
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert(tz).dt.tz_localize(None)
    else:
        return s


def _check_series_convert_timestamps_internal(s, timezone):
    """
    Convert a tz-naive timestamp in the specified timezone or local timezone to UTC normalized for
    Spark internal storage

    Parameters
    ----------
    s : pandas.Series
    timezone : str
        the timezone to convert. if None then use local timezone

    Returns
    -------
    pandas.Series
        `pandas.Series` where if it is a timestamp, has been UTC normalized without a time zone
    """
    from pyspark.sql.pandas.utils import require_minimum_pandas_version
    require_minimum_pandas_version()

    from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64_dtype(s.dtype):
        # When tz_localize a tz-naive timestamp, the result is ambiguous if the tz-naive
        # timestamp is during the hour when the clock is adjusted backward during due to
        # daylight saving time (dst).
        # E.g., for America/New_York, the clock is adjusted backward on 2015-11-01 2:00 to
        # 2015-11-01 1:00 from dst-time to standard time, and therefore, when tz_localize
        # a tz-naive timestamp 2015-11-01 1:30 with America/New_York timezone, it can be either
        # dst time (2015-01-01 1:30-0400) or standard time (2015-11-01 1:30-0500).
        #
        # Here we explicit choose to use standard time. This matches the default behavior of
        # pytz.
        #
        # Here are some code to help understand this behavior:
        # >>> import datetime
        # >>> import pandas as pd
        # >>> import pytz
        # >>>
        # >>> t = datetime.datetime(2015, 11, 1, 1, 30)
        # >>> ts = pd.Series([t])
        # >>> tz = pytz.timezone('America/New_York')
        # >>>
        # >>> ts.dt.tz_localize(tz, ambiguous=True)
        # 0   2015-11-01 01:30:00-04:00
        # dtype: datetime64[ns, America/New_York]
        # >>>
        # >>> ts.dt.tz_localize(tz, ambiguous=False)
        # 0   2015-11-01 01:30:00-05:00
        # dtype: datetime64[ns, America/New_York]
        # >>>
        # >>> str(tz.localize(t))
        # '2015-11-01 01:30:00-05:00'
        tz = timezone or _get_local_timezone()
        return s.dt.tz_localize(tz, ambiguous=False).dt.tz_convert('UTC')
    elif is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert('UTC')
    else:
        return s


def _check_series_convert_timestamps_localize(s, from_timezone, to_timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    Parameters
    ----------
    s : pandas.Series
    from_timezone : str
        the timezone to convert from. if None then use local timezone
    to_timezone : str
        the timezone to convert to. if None then use local timezone

    Returns
    -------
    pandas.Series
        `pandas.Series` where if it is a timestamp, has been converted to tz-naive
    """
    from pyspark.sql.pandas.utils import require_minimum_pandas_version
    require_minimum_pandas_version()

    import pandas as pd
    from pandas.api.types import is_datetime64tz_dtype, is_datetime64_dtype
    from_tz = from_timezone or _get_local_timezone()
    to_tz = to_timezone or _get_local_timezone()
    # TODO: handle nested timestamps, such as ArrayType(TimestampType())?
    if is_datetime64tz_dtype(s.dtype):
        return s.dt.tz_convert(to_tz).dt.tz_localize(None)
    elif is_datetime64_dtype(s.dtype) and from_tz != to_tz:
        # `s.dt.tz_localize('tzlocal()')` doesn't work properly when including NaT.
        return s.apply(
            lambda ts: ts.tz_localize(from_tz, ambiguous=False).tz_convert(to_tz).tz_localize(None)
            if ts is not pd.NaT else pd.NaT)
    else:
        return s


def _check_series_convert_timestamps_local_tz(s, timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    Parameters
    ----------
    s : pandas.Series
    timezone : str
        the timezone to convert to. if None then use local timezone

    Returns
    -------
    pandas.Series
        `pandas.Series` where if it is a timestamp, has been converted to tz-naive
    """
    return _check_series_convert_timestamps_localize(s, None, timezone)


def _check_series_convert_timestamps_tz_local(s, timezone):
    """
    Convert timestamp to timezone-naive in the specified timezone or local timezone

    Parameters
    ----------
    s : pandas.Series
    timezone : str
        the timezone to convert from. if None then use local timezone

    Returns
    -------
    pandas.Series
        `pandas.Series` where if it is a timestamp, has been converted to tz-naive
    """
    return _check_series_convert_timestamps_localize(s, timezone, None)


def _convert_map_items_to_dict(s):
    """
    Convert a series with items as list of (key, value), as made from an Arrow column of map type,
    to dict for compatibility with non-arrow MapType columns.
    :param s: pandas.Series of lists of (key, value) pairs
    :return: pandas.Series of dictionaries
    """
    return s.apply(lambda m: None if m is None else {k: v for k, v in m})


def _convert_dict_to_map_items(s):
    """
    Convert a series of dictionaries to list of (key, value) pairs to match expected data
    for Arrow column of map type.
    :param s: pandas.Series of dictionaries
    :return: pandas.Series of lists of (key, value) pairs
    """
    return s.apply(lambda d: list(d.items()) if d is not None else None)



import logging
from datetime import datetime

from tecton_core import conf
from tecton_core.query.dialect import Dialect
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage


logger = logging.getLogger(__name__)


class QueryTreeMonitor:
    def create_stage(self, type_: TectonManagedStage.StageType, description: str) -> int:
        """Returns stage identifier"""

    def set_query(self, stage_id: int, sql: str) -> None:
        pass

    def update_progress(self, stage_id: int, progress: float) -> None:
        pass

    def set_failed(self, stage_id: int, user_error: bool) -> None:
        pass

    def set_completed(self, stage_id: int) -> None:
        pass

    def set_overall_state(self, state: TectonManagedStage.State) -> None:
        pass


class DebugOutput(QueryTreeMonitor):
    def __init__(self):
        self.start_time = None
        self.step = None
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")

    def create_stage(self, type_: TectonManagedStage.StageType, description: str) -> int:
        self.start_time = datetime.now()
        self.step = description
        if self.is_debug:
            logger.warning(f"------------- Executing stage: {description} -------------")
        return 0

    def set_completed(self, stage_id: int) -> None:
        stage_done_time = datetime.now()
        if self.is_debug:
            logger.warning(f"{self.step} took time (sec): {(stage_done_time - self.start_time).total_seconds()}")

    def set_failed(self, stage_id: int, user_error: bool) -> None:
        return self.set_completed(stage_id)


def get_stage_type_for_dialect(dialect: Dialect) -> TectonManagedStage.StageType:
    if dialect == Dialect.SNOWFLAKE:
        return TectonManagedStage.StageType.SNOWFLAKE
    elif dialect == Dialect.BIGQUERY:
        return TectonManagedStage.StageType.BIGQUERY

    return TectonManagedStage.StageType.PYTHON



from __future__ import annotations

import copy
import enum
import json
import os
import shutil
from dataclasses import dataclass
from functools import lru_cache
from logging import getLogger
from tempfile import TemporaryDirectory
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from urllib.request import urlretrieve

import boto3
from boto3.s3.transfer import S3Transfer
from boto3.s3.transfer import TransferConfig

from tecton_core import conf
from tecton_core.embeddings.file_utils import hash_file


S3_BUCKET = "tecton.ai.public"
AWS_REGION = "us-west-2"
ARCHIVE_FN = "model.tar.gz"
METADATA_FN = "metadata.json"

logger = getLogger(__name__)


class InferenceDeviceHardware(str, enum.Enum):
    CPU = "cpu"
    NVIDIA_T4 = "Tesla T4"
    NVIDIA_M60 = "Tesla M60"
    NVIDIA_V100_16GB = "Tesla V100-SXM2-16GB"
    NVIDIA_A10G = "NVIDIA A10G"
    NVIDIA_L4 = "NVIDIA L4"

    def __str__(self):
        return self.value


class TokenBudgetManager:
    def __init__(self, hardware_token_budget_map: Optional[Dict[InferenceDeviceHardware, int]] = None):  # noqa: ANN204
        if hardware_token_budget_map is None:
            hardware_token_budget_map = {}
        for k, v in hardware_token_budget_map.items():
            if not isinstance(k, InferenceDeviceHardware):
                msg = f"Key {k} is not an instance of InferenceDeviceHardware."
                raise ValueError(msg)
            if not isinstance(v, int) or v <= 0:
                msg = f"Value {v} must be an int greater than 0"
                raise ValueError(msg)
        self._hardware_token_budget_map = copy.deepcopy(hardware_token_budget_map)

    def get(self, hardware: InferenceDeviceHardware) -> Optional[int]:
        return self._hardware_token_budget_map.get(hardware)

    def set(self, hardware: InferenceDeviceHardware, token_budget: int) -> None:
        self._hardware_token_budget_map[hardware] = token_budget


CURRENT_ARTIFACTS = {
    "sentence-transformers/all-MiniLM-L6-v2": (
        "s3://tecton.ai.public/models/huggingface/sentence-transformers/all-MiniLM-L6-v2/e4ce987/1711700006",
        TokenBudgetManager(
            {
                InferenceDeviceHardware.CPU: 1000,
                InferenceDeviceHardware.NVIDIA_T4: 68_768,
                InferenceDeviceHardware.NVIDIA_M60: 12_800,
                InferenceDeviceHardware.NVIDIA_V100_16GB: 204_800,
                InferenceDeviceHardware.NVIDIA_A10G: 150_000,
            }
        ),
    ),
    "mixedbread-ai/mxbai-embed-large-v1": (
        "s3://tecton.ai.public/models/huggingface/mixedbread-ai/mxbai-embed-large-v1/456b7cf/1711697884",
        None,
    ),
    "BAAI/bge-large-en-v1.5": (
        "s3://tecton.ai.public/models/huggingface/BAAI/bge-large-en-v1.5/d4aa690/1711698014",
        None,
    ),
    "BAAI/bge-base-en-v1.5": (
        "s3://tecton.ai.public/models/huggingface/BAAI/bge-base-en-v1.5/a5beb1e/1711698489",
        None,
    ),
    "BAAI/bge-small-en-v1.5": (
        "s3://tecton.ai.public/models/huggingface/BAAI/bge-small-en-v1.5/5c38ec7/1711698601",
        None,
    ),
    "thenlper/gte-large": ("s3://tecton.ai.public/models/huggingface/thenlper/gte-large/5857861/1711699451", None),
    "thenlper/gte-base": ("s3://tecton.ai.public/models/huggingface/thenlper/gte-base/5e95d41/1711699488", None),
    "thenlper/gte-small": ("s3://tecton.ai.public/models/huggingface/thenlper/gte-small/50c7dd3/1711699502", None),
    "Snowflake/snowflake-arctic-embed-xs": (
        "s3://tecton.ai.public/models/huggingface/Snowflake/snowflake-arctic-embed-xs/86a0765/1713945940",
        None,
    ),
    "Snowflake/snowflake-arctic-embed-s": (
        " s3://tecton.ai.public/models/huggingface/Snowflake/snowflake-arctic-embed-s/1d03b2d/1713945950",
        None,
    ),
    "Snowflake/snowflake-arctic-embed-m": (
        "s3://tecton.ai.public/models/huggingface/Snowflake/snowflake-arctic-embed-m/2169d31/1713946149",
        None,
    ),
    "Snowflake/snowflake-arctic-embed-l": (
        "s3://tecton.ai.public/models/huggingface/Snowflake/snowflake-arctic-embed-l/c58efbc/1713946326",
        None,
    ),
    "intfloat/e5-large-unsupervised": (
        "s3://tecton.ai.public/models/huggingface/intfloat/e5-large-unsupervised/15af928/1713946911",
        None,
    ),
    "intfloat/e5-small-unsupervised": (
        "s3://tecton.ai.public/models/huggingface/intfloat/e5-small-unsupervised/b1dcdb2/1713947184",
        None,
    ),
    "intfloat/e5-base-unsupervised": (
        "s3://tecton.ai.public/models/huggingface/intfloat/e5-base-unsupervised/6003a5b/1713947222",
        None,
    ),
}


@dataclass
class ArtifactMetadata:
    model_id: str
    commit_hash: str
    created_at: str
    files: List[str]
    content_hashes: Dict[str, str]
    model_scan_results: Dict[str, Any]

    @staticmethod
    def from_json_file(metadata_file: str) -> ArtifactMetadata:
        with open(metadata_file, "r") as f:
            metadata = json.load(f)
            return ArtifactMetadata(
                model_id=metadata["model_id"],
                commit_hash=metadata["commit_hash"],
                created_at=metadata["created_at"],
                files=metadata["files"],
                content_hashes=metadata["content_hashes"],
                model_scan_results=metadata["model_scan_results"],
            )


class EmbeddingModelInfo:
    def __init__(
        self,
        hf_repo_id: str,
        s3_path: str,
        s3_bucket: str = S3_BUCKET,
        region: str = AWS_REGION,
        token_budget_manager: Optional[TokenBudgetManager] = None,
    ) -> None:
        self.hf_repo_id = hf_repo_id
        self.s3_path = s3_path
        self.s3_bucket = s3_bucket
        self.region = region
        self.__artifact_dir = s3_path.split(S3_BUCKET, 1)[-1]
        self.__http_url_base = f"https://s3.{AWS_REGION}.amazonaws.com/{S3_BUCKET}{self.__artifact_dir}"
        self.metadata_uri = f"{self.s3_path}/{METADATA_FN}"
        self.metadata_object_key = self.object_key_from_uri(self.metadata_uri)
        self.metadata_public_url = f"{self.__http_url_base}/{METADATA_FN}"
        self.model_uri = f"{self.s3_path}/{ARCHIVE_FN}"
        self.model_object_key = self.object_key_from_uri(self.model_uri)
        self.model_public_url = f"{self.__http_url_base}/{ARCHIVE_FN}"
        self.token_budget_manager = token_budget_manager or TokenBudgetManager()

    def get_token_budget(self, device: InferenceDeviceHardware) -> Optional[int]:
        return self.token_budget_manager.get(device)

    def set_token_budget(self, device: InferenceDeviceHardware, token_budget: int) -> None:
        self.token_budget_manager.set(device, token_budget)

    @staticmethod
    def object_key_from_uri(uri: str) -> str:
        return uri.split(S3_BUCKET, 1)[-1][1:]


class ArtifactsProvider:
    def __init__(self, artifact_s3_map: Optional[Dict[str, Tuple[str, TokenBudgetManager]]] = None):  # noqa: ANN204
        artifact_s3_map = CURRENT_ARTIFACTS if artifact_s3_map is None else artifact_s3_map
        self.artifacts = {
            k: EmbeddingModelInfo(k, s3_path, token_budget_manager=token_budget_manager)
            for k, (s3_path, token_budget_manager) in artifact_s3_map.items()
        }

    def get_model_info(self, hf_repo_id: str) -> Optional[EmbeddingModelInfo]:
        return self.artifacts.get(hf_repo_id)


ARTIFACT_PROVIDER = ArtifactsProvider()


def _verify_hashes(metadata: ArtifactMetadata, archive_dir: str) -> None:
    """
    Verify the hash of the model archive.
    """
    for file in metadata.files:
        fp = os.path.join(archive_dir, file)
        if not os.path.exists(fp):
            msg = f"File '{fp}' not found in model directory."
            raise FileNotFoundError(msg)
        actual_hash = hash_file(fp)
        if actual_hash != metadata.content_hashes[file]:
            msg = f"Hash of '{file}' does not match the expected hash."
            raise ValueError(msg)


@lru_cache()
def _get_s3_client():
    return boto3.client("s3")


def retrieve_artifacts(model_name: str, use_http: bool = False) -> str:
    py_model_info = ARTIFACT_PROVIDER.get_model_info(model_name)
    if py_model_info is None:
        msg = f"Model '{model_name}' not found in the artifact provider."
        raise ValueError(msg)
    model_cache_dir = os.path.join(conf.get_or_raise("MODEL_CACHE_DIRECTORY"), model_name)
    if os.path.exists(os.path.join(model_cache_dir, METADATA_FN)):
        logger.info("Model already exists in cache. Skipping download.")
        return model_cache_dir
    os.makedirs(model_cache_dir, exist_ok=True)
    with TemporaryDirectory() as temp_dir:
        metadata_path = os.path.join(temp_dir, METADATA_FN)
        archive_path = os.path.join(temp_dir, ARCHIVE_FN)
        if use_http:
            urlretrieve(py_model_info.metadata_public_url, metadata_path)
            urlretrieve(py_model_info.model_public_url, archive_path)
        else:
            transfer_manager = S3Transfer(
                _get_s3_client(),
                config=TransferConfig(),
            )
            transfer_manager.download_file(py_model_info.s3_bucket, py_model_info.model_object_key, archive_path)
            transfer_manager.download_file(py_model_info.s3_bucket, py_model_info.metadata_object_key, metadata_path)
        metadata = ArtifactMetadata.from_json_file(metadata_path)
        shutil.unpack_archive(archive_path, temp_dir, format="gztar")
        _verify_hashes(metadata, temp_dir)
        for file in metadata.files:
            shutil.move(os.path.join(temp_dir, file), os.path.join(model_cache_dir, file))
        shutil.move(metadata_path, model_cache_dir)
    logger.info(f"Model '{model_name}': {os.listdir(model_cache_dir)} downloaded to '{model_cache_dir}'.")
    return model_cache_dir



import pytest

from tecton import tecton_context
from tecton_core import conf
from tecton_spark.udf_jar import get_udf_jar_path


def pytest_configure():
    # TECTON_FORCE_FUNCTION_SERIALIZATION will always be set during `tecton test`, however it may not be set when
    # running tests via `tecton plan/apply` or when a user invokes tests directly with pytest. In those cases, default
    # TECTON_FORCE_FUNCTION_SERIALIZATION to true.
    if conf.get_or_none("TECTON_FORCE_FUNCTION_SERIALIZATION") is None:
        conf.set("TECTON_FORCE_FUNCTION_SERIALIZATION", "true")


@pytest.fixture(scope="session")
def tecton_pytest_spark_session():
    try:
        from pyspark.sql import SparkSession
    except ImportError:
        pytest.fail("Cannot create a SparkSession if `pyspark` is not installed.")

    active_session = SparkSession.getActiveSession()
    if active_session:
        pytest.fail(
            f"Cannot create SparkSession `tecton_pytest_spark_session` when there is already an active session: {active_session.sparkContext.appName}"
        )

    try:
        spark = (
            SparkSession.builder.appName("tecton_pytest_spark_session")
            .config("spark.jars", get_udf_jar_path())
            # This short-circuit's Spark's attempt to auto-detect a hostname for the master address, which can lead to
            # errors on hosts with "unusual" hostnames that Spark believes are invalid.
            .config("spark.driver.host", "localhost")
            .config("spark.ui.enabled", "false")
            .config("spark.sql.session.timeZone", "UTC")
            # Enable Arrow for PySpark to be compatible with Pandas 2.0. More details, refer to:
            # https://www.notion.so/tecton/mini-RFC-Pandas-2-0-Spark-Compatibility-70687042a6144e60b0bc2a622af65f09?pvs=4#ef8b77bdcc8e4147a224543725ea1936
            .config("spark.sql.execution.arrow.pyspark.enabled", "true")
            .getOrCreate()
        )
    except Exception as e:
        # Unfortunately we can't do much better than parsing the error message since Spark raises `Exception` rather than a more specific type.
        if str(e) == "Java gateway process exited before sending its port number":
            pytest.fail(
                "Failed to start Java process for Spark, perhaps Java isn't installed or the 'JAVA_HOME' environment variable is not set?"
            )

        raise e

    try:
        tecton_context.set_tecton_spark_session(spark)
        yield spark
    finally:
        spark.stop()



from dataclasses import dataclass
from typing import List

import pytest

from tecton.cli.environment import DEFAULT_UPLOAD_PART_SIZE_MB
from tecton.cli.environment import UploadPart
from tecton.cli.environment import get_upload_parts
from tecton.cli.environment_utils import EnvironmentDependencies
from tecton.cli.environment_utils import is_valid_environment_name


SAMPLE_LOCK_JSON = {
    "locked_resolves": [
        {
            "locked_requirements": [
                {
                    "project_name": "attrs",
                    "requires_dists": [
                        "attrs[tests-mypy]; extra == 'tests-no-zope'",
                        "hypothesis; extra == 'tests-no-zope'",
                    ],
                    "version": "23.2.0",
                },
                {"project_name": "boto3", "requires_dists": ["botocore<1.35.0,>=1.34.87"], "version": "1.34.87"},
                {"project_name": "pytest", "requires_dists": [], "version": "8.1.1"},
            ]
        }
    ],
    "requirements": ["attrs==23.2.0", "boto3==1.34.87", "pytest"],
}


@dataclass
class FileSplit__TestCase:
    name: str
    file_size: int
    expected_parts: List[UploadPart]


FILE_SPLIT_TEST_CASES = [
    FileSplit__TestCase(
        name="single_file",
        file_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 - 1,
        expected_parts=[UploadPart(part_number=1, offset=0, part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 - 1)],
    ),
    FileSplit__TestCase(
        name="exact_multiple_parts",
        file_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 * 5,
        expected_parts=[
            UploadPart(
                part_number=i,
                offset=(i - 1) * DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
            )
            for i in range(1, 6)
        ],
    ),
    FileSplit__TestCase(
        name="multiple_parts_with_last_part_smaller",
        file_size=(DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 * 2) + (DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 // 2),
        expected_parts=[
            UploadPart(part_number=1, offset=0, part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024),
            UploadPart(
                part_number=2,
                offset=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
            ),
            UploadPart(
                part_number=3,
                offset=2 * DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024,
                part_size=DEFAULT_UPLOAD_PART_SIZE_MB * 1024 * 1024 // 2,
            ),
        ],
    ),
    FileSplit__TestCase(
        name="zero_size_file",
        file_size=0,
        expected_parts=[],
    ),
]


@pytest.fixture
def temporary_requirements_file(tmp_path):
    requirements_text = "package1\n# package2\npackage3\n"
    requirements_path = tmp_path / "requirements.txt"
    requirements_path.write_text(requirements_text)
    return requirements_path


@pytest.mark.parametrize("test_case", FILE_SPLIT_TEST_CASES, ids=[tc.name for tc in FILE_SPLIT_TEST_CASES])
def test_get_upload_parts(test_case):
    parts = get_upload_parts(test_case.file_size)
    assert len(parts) == len(test_case.expected_parts)
    for part, expected_part in zip(parts, test_case.expected_parts):
        assert part.part_size == expected_part.part_size
        assert part.part_number == expected_part.part_number
        assert part.offset == expected_part.offset


@pytest.mark.parametrize(
    "name, expected",
    [
        ("env123", True),
        ("env_123", True),
        ("ENV-123", True),
        ("env*123", False),
        ("env?123", False),
        ("env!123", False),
        ("", False),
        ("env 123", False),
    ],
)
def test_environments(name, expected):
    assert is_valid_environment_name(name) == expected


@pytest.fixture
def environment_dependencies():
    return EnvironmentDependencies(SAMPLE_LOCK_JSON)


@pytest.mark.parametrize(
    "package_name, expected_version",
    [("attrs", "23.2.0"), ("boto3", "1.34.87"), ("pytest", "8.1.1"), ("nonexistent", None)],
)
def test_get_version(environment_dependencies, package_name, expected_version):
    assert environment_dependencies.get_version(package_name) == expected_version


@pytest.mark.parametrize(
    "package_name, expected_presence", [("attrs", True), ("boto3", True), ("pytest", True), ("nonexistent", False)]
)
def test_is_dependency_present(environment_dependencies, package_name, expected_presence):
    assert environment_dependencies.is_dependency_present(package_name) == expected_presence


@pytest.mark.parametrize(
    "package_name, expected_extras", [("attrs", ["tests-no-zope"]), ("boto3", []), ("pytest", []), ("nonexistent", [])]
)
def test_get_dependency_extras(environment_dependencies, package_name, expected_extras):
    extras = environment_dependencies.get_dependency_extras(package_name)
    assert extras == expected_extras



from enum import Enum
from typing import Optional
from typing import Union

from tecton_core import conf
from tecton_core.errors import TectonValidationError
from tecton_core.query.dialect import Dialect
from tecton_proto.common import compute_mode__client_pb2 as compute_mode_pb2


class ComputeMode(str, Enum):
    """Represents the compute mode for training data generation queries."""

    SPARK = "spark"
    SNOWFLAKE = "snowflake"
    ATHENA = "athena"
    RIFT = "rift"

    def default_dialect(self) -> Dialect:
        return _COMPUTE_MODE_TO_DIALECT[self]


_COMPUTE_MODE_TO_DIALECT = {
    ComputeMode.SPARK: Dialect.SPARK,
    ComputeMode.SNOWFLAKE: Dialect.SNOWFLAKE,
    ComputeMode.ATHENA: Dialect.ATHENA,
    ComputeMode.RIFT: Dialect.DUCKDB,
}


class ComputeModeValueError(ValueError):
    def __init__(self, compute_mode_str: str) -> None:
        msg = f"Invalid compute mode: {compute_mode_str}. Must be one of {[e.value for e in ComputeMode]}"
        super().__init__(msg)


def _parse_compute_mode(compute_mode_str: str) -> ComputeMode:
    try:
        return ComputeMode(compute_mode_str.lower())
    except ValueError:
        raise ComputeModeValueError(compute_mode_str)


def _default_offline_retrieval_compute_mode() -> ComputeMode:
    return _parse_compute_mode(conf.get_or_raise("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE"))


def offline_retrieval_compute_mode(raw_function_override: Optional[Union[ComputeMode, str]]) -> ComputeMode:
    """Returns offline retrieval compute mode based on the given function-level override and the environment."""
    default = _default_offline_retrieval_compute_mode()
    if raw_function_override is not None:
        function_override = _parse_compute_mode(raw_function_override)
        if function_override == ComputeMode.SNOWFLAKE and default != ComputeMode.SNOWFLAKE:
            raise TectonValidationError(
                "Snowflake offline retrieval cannot be enabled at a function level; it must be enabled globally. "
                + "Try setting TECTON_OFFLINE_RETRIEVAL_MODE to snowflake in your environment variables, "
                + 'or running tecton.conf.set("TECTON_OFFLINE_RETRIEVAL_MODE", "snowflake"))'
            )
        if function_override != ComputeMode.SNOWFLAKE and default == ComputeMode.SNOWFLAKE:
            raise TectonValidationError(
                "Snowflake offline retrieval cannot be overridden at a function level when it is enabled globally. "
                + f"Try setting TECTON_OFFLINE_RETRIEVAL_MODE to {function_override.value} in your environment variables, "
                + f'or running tecton.conf.set("TECTON_OFFLINE_RETRIEVAL_MODE", "{function_override.value}"))'
            )
        return function_override
    else:
        return default


class BatchComputeMode(Enum):
    """Represents that compute mode for batch jobs associated with a FeatureView."""

    SPARK = compute_mode_pb2.BatchComputeMode.BATCH_COMPUTE_MODE_SPARK
    SNOWFLAKE = compute_mode_pb2.BatchComputeMode.BATCH_COMPUTE_MODE_SNOWFLAKE
    RIFT = compute_mode_pb2.BatchComputeMode.BATCH_COMPUTE_MODE_RIFT

    @property
    def value(self) -> compute_mode_pb2.BatchComputeMode:
        return super().value


_STR_TO_BATCH = {
    "spark": BatchComputeMode.SPARK,
    "snowflake": BatchComputeMode.SNOWFLAKE,
    "rift": BatchComputeMode.RIFT,
}


def default_batch_compute_mode() -> BatchComputeMode:
    return _STR_TO_BATCH[conf.get_or_raise("TECTON_BATCH_COMPUTE_MODE").lower()]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/offlinestore/delta/metadata__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6tecton_proto/offlinestore/delta/metadata__client.proto\x12\x1ftecton_proto.offlinestore.delta\x1a\x1fgoogle/protobuf/timestamp.proto\"\xb7\x01\n\x13TectonDeltaMetadata\x12%\n\rdeletion_path\x18\x01 \x01(\tH\x00R\x0c\x64\x65letionPath\x12!\n\x0bingest_path\x18\x02 \x01(\tH\x00R\ningestPath\x12J\n\x12\x66\x65\x61ture_start_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\x10\x66\x65\x61tureStartTimeB\n\n\x08key_typeB!\n\x1d\x63om.tecton.offlinestore.deltaP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.offlinestore.delta.metadata__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.tecton.offlinestore.deltaP\001'
  _TECTONDELTAMETADATA._serialized_start=125
  _TECTONDELTAMETADATA._serialized_end=308
# @@protoc_insertion_point(module_scope)



import logging
import uuid
from typing import Dict

import pyarrow

from tecton_core.query import nodes
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.pandas import nodes as pandas_nodes
from tecton_core.query.query_tree_compute import QueryTreeCompute


logger = logging.getLogger(__name__)


class PandasTreeRewriter:
    def rewrite(
        self,
        tree: NodeRef,
        pipeline_compute: QueryTreeCompute,
        data_sources: Dict[str, pyarrow.RecordBatchReader],
    ) -> None:
        """Finds all FeatureViewPipelineNodes, executes their subtrees, and replaces them with StagedTableScanNodes.

        Assumes that the inputs to the FeatureViewPipelineNodes have already been replaced with StagedTableScanNodes,
        and that each such StagedTableScanNode corresponds to a pyarrow table contained in 'data_sources'.
        """
        tree_node = tree.node

        if isinstance(tree_node, nodes.FeatureViewPipelineNode):
            for _, fv_input_node_ref in tree_node.inputs_map.items():
                self._rewrite_fv_input_node(fv_input_node_ref, data_sources)

            pipeline_node = pandas_nodes.PandasFeatureViewPipelineNode.from_node_inputs(
                query_node=tree_node,
                input_node=None,
            )
            pipeline_result = pipeline_node.to_arrow()
            staging_table_name = f"{pipeline_node.feature_definition_wrapper.name}_{uuid.uuid4().hex[:16]}_pandas"
            tree.node = nodes.StagedTableScanNode(
                tree_node.dialect,
                tree_node.compute_mode,
                staged_columns=pipeline_node.columns,
                staging_table_name=staging_table_name,
            )
            pipeline_compute.register_temp_table(staging_table_name, pipeline_result)
        else:
            for i in tree.inputs:
                self.rewrite(tree=i, pipeline_compute=pipeline_compute, data_sources=data_sources)

    def _rewrite_fv_input_node(self, tree: NodeRef, data_sources: Dict[str, pyarrow.RecordBatchReader]) -> None:
        # Certain StagedTableScanNodes are duplicated. If one of them has already been converted to a PandasDataNode,
        # the other does not need to be converted.
        if isinstance(tree.node, pandas_nodes.PandasDataNode):
            return
        assert isinstance(tree.node, nodes.StagedTableScanNode)
        table_name = tree.node.staging_table_name
        assert table_name in data_sources
        # A rewrite should only leave NodeRefs. However, this PandasDataNode is temporary. It will be removed above.
        tree.node = pandas_nodes.PandasDataNode(
            input_reader=data_sources[table_name],
            input_node=None,
            columns=tree.node.columns,
            column_name_updater=lambda x: x,
            secret_resolver=None,
        )



from abc import ABC
from abc import abstractmethod

from tecton_core.query.node_interface import NodeRef


class Rewrite(ABC):
    @abstractmethod
    def rewrite(self, node: NodeRef) -> None:
        raise NotImplementedError



from types import MappingProxyType
from typing import Mapping
from typing import Tuple

from typeguard import typechecked

from tecton_core import data_types
from tecton_core import schema
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_proto.args import entity__client_pb2 as entity__args_pb2
from tecton_proto.data import entity__client_pb2 as entity__data_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "EntitySpec",
]


@utils.frozen_strict
class EntitySpec(tecton_object_spec.TectonObjectSpec):
    join_keys: Tuple[schema.Column, ...]
    options: Mapping[str, str]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: entity__data_pb2.Entity) -> "EntitySpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(proto.entity_id, proto.fco_metadata),
            join_keys=tuple(
                schema.Column(name=join_key.name, dtype=data_types.data_type_from_proto(join_key.offline_data_type))
                for join_key in proto.join_keys
            ),
            validation_args=validator_pb2.FcoValidationArgs(entity=proto.validation_args),
            options=MappingProxyType(proto.options),
        )

    @classmethod
    @typechecked
    def from_args_proto(cls, proto: entity__args_pb2.EntityArgs) -> "EntitySpec":
        join_keys = tuple(
            schema.Column(name=join_key.name, dtype=data_types.data_type_from_proto(join_key.offline_data_type))
            for join_key in proto.join_keys
        )
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.entity_id, proto.info),
            join_keys=join_keys,
            validation_args=None,
            options=MappingProxyType(proto.options),
        )

    @property
    def join_key_names(self) -> Tuple[str, ...]:
        return tuple(join_key.name for join_key in self.join_keys)



class NodePropertyError(Exception):
    """Basic Node attribute error"""
    pass


class NodeIDAbsentError(NodePropertyError):
    """Exception throwed if a node's identifier is unknown"""
    pass


class NodePropertyAbsentError(NodePropertyError):
    """Exception throwed if a node's data property is not specified"""
    pass


class MultipleRootError(Exception):
    """Exception throwed if more than one root exists in a tree."""
    pass


class DuplicatedNodeIdError(Exception):
    """Exception throwed if an identifier already exists in a tree."""
    pass


class LinkPastRootNodeError(Exception):
    """
    Exception throwed in Tree.link_past_node() if one attempts
    to "link past" the root node of a tree.
    """
    pass


class InvalidLevelNumber(Exception):
    pass


class LoopError(Exception):
    """
    Exception thrown if trying to move node B to node A's position
    while A is B's ancestor.
    """
    pass





# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/data_source_type__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/common/data_source_type__client.proto\x12\x13tecton_proto.common*g\n\x0e\x44\x61taSourceType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05\x42\x41TCH\x10\x01\x12\x15\n\x11STREAM_WITH_BATCH\x10\x02\x12\x11\n\rPUSH_NO_BATCH\x10\x03\x12\x13\n\x0fPUSH_WITH_BATCH\x10\x04\x42?\n\x11\x63om.tecton.commonP\x01Z(github.com/tecton-ai/tecton_proto/common')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.data_source_type__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001Z(github.com/tecton-ai/tecton_proto/common'
  _DATASOURCETYPE._serialized_start=75
  _DATASOURCETYPE._serialized_end=178
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from math import exp
import sys
import warnings

import numpy

from pyspark import RDD, since
from pyspark.mllib.common import callMLlibFunc, _py2java, _java2py
from pyspark.mllib.linalg import _convert_to_vector
from pyspark.mllib.regression import (
    LabeledPoint, LinearModel, _regression_train_wrapper,
    StreamingLinearAlgorithm)
from pyspark.mllib.util import Saveable, Loader, inherit_doc


__all__ = ['LogisticRegressionModel', 'LogisticRegressionWithSGD', 'LogisticRegressionWithLBFGS',
           'SVMModel', 'SVMWithSGD', 'NaiveBayesModel', 'NaiveBayes',
           'StreamingLogisticRegressionWithSGD']


class LinearClassificationModel(LinearModel):
    """
    A private abstract class representing a multiclass classification
    model. The categories are represented by int values: 0, 1, 2, etc.
    """
    def __init__(self, weights, intercept):
        super(LinearClassificationModel, self).__init__(weights, intercept)
        self._threshold = None

    @since('1.4.0')
    def setThreshold(self, value):
        """
        Sets the threshold that separates positive predictions from
        negative predictions. An example with prediction score greater
        than or equal to this threshold is identified as a positive,
        and negative otherwise. It is used for binary classification
        only.
        """
        self._threshold = value

    @property
    @since('1.4.0')
    def threshold(self):
        """
        Returns the threshold (if any) used for converting raw
        prediction scores into 0/1 predictions. It is used for
        binary classification only.
        """
        return self._threshold

    @since('1.4.0')
    def clearThreshold(self):
        """
        Clears the threshold so that `predict` will output raw
        prediction scores. It is used for binary classification only.
        """
        self._threshold = None

    @since('1.4.0')
    def predict(self, test):
        """
        Predict values for a single data point or an RDD of points
        using the model trained.
        """
        raise NotImplementedError


class LogisticRegressionModel(LinearClassificationModel):

    """
    Classification model trained using Multinomial/Binary Logistic
    Regression.

    .. versionadded:: 0.9.0

    Parameters
    ----------
    weights : :py:class:`pyspark.mllib.linalg.Vector`
        Weights computed for every feature.
    intercept : float
        Intercept computed for this model. (Only used in Binary Logistic
        Regression. In Multinomial Logistic Regression, the intercepts will
        not be a single value, so the intercepts will be part of the
        weights.)
    numFeatures : int
        The dimension of the features.
    numClasses : int
        The number of possible outcomes for k classes classification problem
        in Multinomial Logistic Regression. By default, it is binary
        logistic regression so numClasses will be set to 2.

    Examples
    --------
    >>> from pyspark.mllib.linalg import SparseVector
    >>> data = [
    ...     LabeledPoint(0.0, [0.0, 1.0]),
    ...     LabeledPoint(1.0, [1.0, 0.0]),
    ... ]
    >>> lrm = LogisticRegressionWithSGD.train(sc.parallelize(data), iterations=10)
    >>> lrm.predict([1.0, 0.0])
    1
    >>> lrm.predict([0.0, 1.0])
    0
    >>> lrm.predict(sc.parallelize([[1.0, 0.0], [0.0, 1.0]])).collect()
    [1, 0]
    >>> lrm.clearThreshold()
    >>> lrm.predict([0.0, 1.0])
    0.279...

    >>> sparse_data = [
    ...     LabeledPoint(0.0, SparseVector(2, {0: 0.0})),
    ...     LabeledPoint(1.0, SparseVector(2, {1: 1.0})),
    ...     LabeledPoint(0.0, SparseVector(2, {0: 1.0})),
    ...     LabeledPoint(1.0, SparseVector(2, {1: 2.0}))
    ... ]
    >>> lrm = LogisticRegressionWithSGD.train(sc.parallelize(sparse_data), iterations=10)
    >>> lrm.predict(numpy.array([0.0, 1.0]))
    1
    >>> lrm.predict(numpy.array([1.0, 0.0]))
    0
    >>> lrm.predict(SparseVector(2, {1: 1.0}))
    1
    >>> lrm.predict(SparseVector(2, {0: 1.0}))
    0
    >>> import os, tempfile
    >>> path = tempfile.mkdtemp()
    >>> lrm.save(sc, path)
    >>> sameModel = LogisticRegressionModel.load(sc, path)
    >>> sameModel.predict(numpy.array([0.0, 1.0]))
    1
    >>> sameModel.predict(SparseVector(2, {0: 1.0}))
    0
    >>> from shutil import rmtree
    >>> try:
    ...    rmtree(path)
    ... except:
    ...    pass
    >>> multi_class_data = [
    ...     LabeledPoint(0.0, [0.0, 1.0, 0.0]),
    ...     LabeledPoint(1.0, [1.0, 0.0, 0.0]),
    ...     LabeledPoint(2.0, [0.0, 0.0, 1.0])
    ... ]
    >>> data = sc.parallelize(multi_class_data)
    >>> mcm = LogisticRegressionWithLBFGS.train(data, iterations=10, numClasses=3)
    >>> mcm.predict([0.0, 0.5, 0.0])
    0
    >>> mcm.predict([0.8, 0.0, 0.0])
    1
    >>> mcm.predict([0.0, 0.0, 0.3])
    2
    """
    def __init__(self, weights, intercept, numFeatures, numClasses):
        super(LogisticRegressionModel, self).__init__(weights, intercept)
        self._numFeatures = int(numFeatures)
        self._numClasses = int(numClasses)
        self._threshold = 0.5
        if self._numClasses == 2:
            self._dataWithBiasSize = None
            self._weightsMatrix = None
        else:
            self._dataWithBiasSize = self._coeff.size // (self._numClasses - 1)
            self._weightsMatrix = self._coeff.toArray().reshape(self._numClasses - 1,
                                                                self._dataWithBiasSize)

    @property
    @since('1.4.0')
    def numFeatures(self):
        """
        Dimension of the features.
        """
        return self._numFeatures

    @property
    @since('1.4.0')
    def numClasses(self):
        """
        Number of possible outcomes for k classes classification problem
        in Multinomial Logistic Regression.
        """
        return self._numClasses

    @since('0.9.0')
    def predict(self, x):
        """
        Predict values for a single data point or an RDD of points
        using the model trained.
        """
        if isinstance(x, RDD):
            return x.map(lambda v: self.predict(v))

        x = _convert_to_vector(x)
        if self.numClasses == 2:
            margin = self.weights.dot(x) + self._intercept
            if margin > 0:
                prob = 1 / (1 + exp(-margin))
            else:
                exp_margin = exp(margin)
                prob = exp_margin / (1 + exp_margin)
            if self._threshold is None:
                return prob
            else:
                return 1 if prob > self._threshold else 0
        else:
            best_class = 0
            max_margin = 0.0
            if x.size + 1 == self._dataWithBiasSize:
                for i in range(0, self._numClasses - 1):
                    margin = x.dot(self._weightsMatrix[i][0:x.size]) + \
                        self._weightsMatrix[i][x.size]
                    if margin > max_margin:
                        max_margin = margin
                        best_class = i + 1
            else:
                for i in range(0, self._numClasses - 1):
                    margin = x.dot(self._weightsMatrix[i])
                    if margin > max_margin:
                        max_margin = margin
                        best_class = i + 1
            return best_class

    @since('1.4.0')
    def save(self, sc, path):
        """
        Save this model to the given path.
        """
        java_model = sc._jvm.org.apache.spark.mllib.classification.LogisticRegressionModel(
            _py2java(sc, self._coeff), self.intercept, self.numFeatures, self.numClasses)
        java_model.save(sc._jsc.sc(), path)

    @classmethod
    @since('1.4.0')
    def load(cls, sc, path):
        """
        Load a model from the given path.
        """
        java_model = sc._jvm.org.apache.spark.mllib.classification.LogisticRegressionModel.load(
            sc._jsc.sc(), path)
        weights = _java2py(sc, java_model.weights())
        intercept = java_model.intercept()
        numFeatures = java_model.numFeatures()
        numClasses = java_model.numClasses()
        threshold = java_model.getThreshold().get()
        model = LogisticRegressionModel(weights, intercept, numFeatures, numClasses)
        model.setThreshold(threshold)
        return model

    def __repr__(self):
        return self._call_java("toString")


class LogisticRegressionWithSGD(object):
    """
    Train a classification model for Binary Logistic Regression using Stochastic Gradient Descent.

    .. versionadded:: 0.9.0
    .. deprecated:: 2.0.0
        Use ml.classification.LogisticRegression or LogisticRegressionWithLBFGS.
    """
    @classmethod
    def train(cls, data, iterations=100, step=1.0, miniBatchFraction=1.0,
              initialWeights=None, regParam=0.01, regType="l2", intercept=False,
              validateData=True, convergenceTol=0.001):
        """
        Train a logistic regression model on the given data.

        .. versionadded:: 0.9.0

        Parameters
        ----------
        data : :py:class:`pyspark.RDD`
            The training data, an RDD of :py:class:`pyspark.mllib.regression.LabeledPoint`.
        iterations : int, optional
            The number of iterations.
            (default: 100)
        step : float, optional
            The step parameter used in SGD.
            (default: 1.0)
        miniBatchFraction : float, optional
            Fraction of data to be used for each SGD iteration.
            (default: 1.0)
        initialWeights : :py:class:`pyspark.mllib.linalg.Vector` or convertible, optional
            The initial weights.
            (default: None)
        regParam : float, optional
            The regularizer parameter.
            (default: 0.01)
        regType : str, optional
            The type of regularizer used for training our model.
            Supported values:

            - "l1" for using L1 regularization
            - "l2" for using L2 regularization (default)
            - None for no regularization

        intercept : bool, optional
            Boolean parameter which indicates the use or not of the
            augmented representation for training data (i.e., whether bias
            features are activated or not).
            (default: False)
        validateData : bool, optional
            Boolean parameter which indicates if the algorithm should
            validate data before training.
            (default: True)
        convergenceTol : float, optional
            A condition which decides iteration termination.
            (default: 0.001)
        """
        warnings.warn(
            "Deprecated in 2.0.0. Use ml.classification.LogisticRegression or "
            "LogisticRegressionWithLBFGS.", DeprecationWarning)

        def train(rdd, i):
            return callMLlibFunc("trainLogisticRegressionModelWithSGD", rdd, int(iterations),
                                 float(step), float(miniBatchFraction), i, float(regParam), regType,
                                 bool(intercept), bool(validateData), float(convergenceTol))

        return _regression_train_wrapper(train, LogisticRegressionModel, data, initialWeights)


class LogisticRegressionWithLBFGS(object):
    """
    Train a classification model for Multinomial/Binary Logistic Regression
    using Limited-memory BFGS.

    Standard feature scaling and L2 regularization are used by default.
    .. versionadded:: 1.2.0
    """
    @classmethod
    def train(cls, data, iterations=100, initialWeights=None, regParam=0.0, regType="l2",
              intercept=False, corrections=10, tolerance=1e-6, validateData=True, numClasses=2):
        """
        Train a logistic regression model on the given data.

        .. versionadded:: 1.2.0

        Parameters
        ----------
        data : :py:class:`pyspark.RDD`
            The training data, an RDD of :py:class:`pyspark.mllib.regression.LabeledPoint`.
        iterations : int, optional
            The number of iterations.
            (default: 100)
        initialWeights : :py:class:`pyspark.mllib.linalg.Vector` or convertible, optional
            The initial weights.
            (default: None)
        regParam : float, optional
            The regularizer parameter.
            (default: 0.01)
        regType : str, optional
            The type of regularizer used for training our model.
            Supported values:

            - "l1" for using L1 regularization
            - "l2" for using L2 regularization (default)
            - None for no regularization

        intercept : bool, optional
            Boolean parameter which indicates the use or not of the
            augmented representation for training data (i.e., whether bias
            features are activated or not).
            (default: False)
        corrections : int, optional
            The number of corrections used in the LBFGS update.
            If a known updater is used for binary classification,
            it calls the ml implementation and this parameter will
            have no effect. (default: 10)
        tolerance : float, optional
            The convergence tolerance of iterations for L-BFGS.
            (default: 1e-6)
        validateData : bool, optional
            Boolean parameter which indicates if the algorithm should
            validate data before training.
            (default: True)
        numClasses : int, optional
            The number of classes (i.e., outcomes) a label can take in
            Multinomial Logistic Regression.
            (default: 2)

        Examples
        --------
        >>> data = [
        ...     LabeledPoint(0.0, [0.0, 1.0]),
        ...     LabeledPoint(1.0, [1.0, 0.0]),
        ... ]
        >>> lrm = LogisticRegressionWithLBFGS.train(sc.parallelize(data), iterations=10)
        >>> lrm.predict([1.0, 0.0])
        1
        >>> lrm.predict([0.0, 1.0])
        0
        """
        def train(rdd, i):
            return callMLlibFunc("trainLogisticRegressionModelWithLBFGS", rdd, int(iterations), i,
                                 float(regParam), regType, bool(intercept), int(corrections),
                                 float(tolerance), bool(validateData), int(numClasses))

        if initialWeights is None:
            if numClasses == 2:
                initialWeights = [0.0] * len(data.first().features)
            else:
                if intercept:
                    initialWeights = [0.0] * (len(data.first().features) + 1) * (numClasses - 1)
                else:
                    initialWeights = [0.0] * len(data.first().features) * (numClasses - 1)
        return _regression_train_wrapper(train, LogisticRegressionModel, data, initialWeights)


class SVMModel(LinearClassificationModel):

    """
    Model for Support Vector Machines (SVMs).

    .. versionadded:: 0.9.0

    Parameters
    ----------
    weights : :py:class:`pyspark.mllib.linalg.Vector`
        Weights computed for every feature.
    intercept : float
        Intercept computed for this model.

    Examples
    --------
    >>> from pyspark.mllib.linalg import SparseVector
    >>> data = [
    ...     LabeledPoint(0.0, [0.0]),
    ...     LabeledPoint(1.0, [1.0]),
    ...     LabeledPoint(1.0, [2.0]),
    ...     LabeledPoint(1.0, [3.0])
    ... ]
    >>> svm = SVMWithSGD.train(sc.parallelize(data), iterations=10)
    >>> svm.predict([1.0])
    1
    >>> svm.predict(sc.parallelize([[1.0]])).collect()
    [1]
    >>> svm.clearThreshold()
    >>> svm.predict(numpy.array([1.0]))
    1.44...

    >>> sparse_data = [
    ...     LabeledPoint(0.0, SparseVector(2, {0: -1.0})),
    ...     LabeledPoint(1.0, SparseVector(2, {1: 1.0})),
    ...     LabeledPoint(0.0, SparseVector(2, {0: 0.0})),
    ...     LabeledPoint(1.0, SparseVector(2, {1: 2.0}))
    ... ]
    >>> svm = SVMWithSGD.train(sc.parallelize(sparse_data), iterations=10)
    >>> svm.predict(SparseVector(2, {1: 1.0}))
    1
    >>> svm.predict(SparseVector(2, {0: -1.0}))
    0
    >>> import os, tempfile
    >>> path = tempfile.mkdtemp()
    >>> svm.save(sc, path)
    >>> sameModel = SVMModel.load(sc, path)
    >>> sameModel.predict(SparseVector(2, {1: 1.0}))
    1
    >>> sameModel.predict(SparseVector(2, {0: -1.0}))
    0
    >>> from shutil import rmtree
    >>> try:
    ...    rmtree(path)
    ... except:
    ...    pass
    """
    def __init__(self, weights, intercept):
        super(SVMModel, self).__init__(weights, intercept)
        self._threshold = 0.0

    @since('0.9.0')
    def predict(self, x):
        """
        Predict values for a single data point or an RDD of points
        using the model trained.
        """
        if isinstance(x, RDD):
            return x.map(lambda v: self.predict(v))

        x = _convert_to_vector(x)
        margin = self.weights.dot(x) + self.intercept
        if self._threshold is None:
            return margin
        else:
            return 1 if margin > self._threshold else 0

    @since('1.4.0')
    def save(self, sc, path):
        """
        Save this model to the given path.
        """
        java_model = sc._jvm.org.apache.spark.mllib.classification.SVMModel(
            _py2java(sc, self._coeff), self.intercept)
        java_model.save(sc._jsc.sc(), path)

    @classmethod
    @since('1.4.0')
    def load(cls, sc, path):
        """
        Load a model from the given path.
        """
        java_model = sc._jvm.org.apache.spark.mllib.classification.SVMModel.load(
            sc._jsc.sc(), path)
        weights = _java2py(sc, java_model.weights())
        intercept = java_model.intercept()
        threshold = java_model.getThreshold().get()
        model = SVMModel(weights, intercept)
        model.setThreshold(threshold)
        return model


class SVMWithSGD(object):
    """
    Train a Support Vector Machine (SVM) using Stochastic Gradient Descent.

    .. versionadded:: 0.9.0
    """

    @classmethod
    def train(cls, data, iterations=100, step=1.0, regParam=0.01,
              miniBatchFraction=1.0, initialWeights=None, regType="l2",
              intercept=False, validateData=True, convergenceTol=0.001):
        """
        Train a support vector machine on the given data.

        .. versionadded:: 0.9.0

        Parameters
        ----------
        data : :py:class:`pyspark.RDD`
            The training data, an RDD of :py:class:`pyspark.mllib.regression.LabeledPoint`.
        iterations : int, optional
            The number of iterations.
            (default: 100)
        step : float, optional
            The step parameter used in SGD.
            (default: 1.0)
        regParam : float, optional
            The regularizer parameter.
            (default: 0.01)
        miniBatchFraction : float, optional
            Fraction of data to be used for each SGD iteration.
            (default: 1.0)
        initialWeights : :py:class:`pyspark.mllib.linalg.Vector` or convertible, optional
            The initial weights.
            (default: None)
        regType : str, optional
            The type of regularizer used for training our model.
            Allowed values:

            - "l1" for using L1 regularization
            - "l2" for using L2 regularization (default)
            - None for no regularization

        intercept : bool, optional
            Boolean parameter which indicates the use or not of the
            augmented representation for training data (i.e. whether bias
            features are activated or not).
            (default: False)
        validateData : bool, optional
            Boolean parameter which indicates if the algorithm should
            validate data before training.
            (default: True)
        convergenceTol : float, optional
            A condition which decides iteration termination.
            (default: 0.001)
        """
        def train(rdd, i):
            return callMLlibFunc("trainSVMModelWithSGD", rdd, int(iterations), float(step),
                                 float(regParam), float(miniBatchFraction), i, regType,
                                 bool(intercept), bool(validateData), float(convergenceTol))

        return _regression_train_wrapper(train, SVMModel, data, initialWeights)


@inherit_doc
class NaiveBayesModel(Saveable, Loader):

    """
    Model for Naive Bayes classifiers.

    .. versionadded:: 0.9.0

    Parameters
    ----------
    labels : :py:class:`numpy.ndarray`
        List of labels.
    pi : :py:class:`numpy.ndarray`
        Log of class priors, whose dimension is C, number of labels.
    theta : :py:class:`numpy.ndarray`
        Log of class conditional probabilities, whose dimension is C-by-D,
        where D is number of features.

    Examples
    --------
    >>> from pyspark.mllib.linalg import SparseVector
    >>> data = [
    ...     LabeledPoint(0.0, [0.0, 0.0]),
    ...     LabeledPoint(0.0, [0.0, 1.0]),
    ...     LabeledPoint(1.0, [1.0, 0.0]),
    ... ]
    >>> model = NaiveBayes.train(sc.parallelize(data))
    >>> model.predict(numpy.array([0.0, 1.0]))
    0.0
    >>> model.predict(numpy.array([1.0, 0.0]))
    1.0
    >>> model.predict(sc.parallelize([[1.0, 0.0]])).collect()
    [1.0]
    >>> sparse_data = [
    ...     LabeledPoint(0.0, SparseVector(2, {1: 0.0})),
    ...     LabeledPoint(0.0, SparseVector(2, {1: 1.0})),
    ...     LabeledPoint(1.0, SparseVector(2, {0: 1.0}))
    ... ]
    >>> model = NaiveBayes.train(sc.parallelize(sparse_data))
    >>> model.predict(SparseVector(2, {1: 1.0}))
    0.0
    >>> model.predict(SparseVector(2, {0: 1.0}))
    1.0
    >>> import os, tempfile
    >>> path = tempfile.mkdtemp()
    >>> model.save(sc, path)
    >>> sameModel = NaiveBayesModel.load(sc, path)
    >>> sameModel.predict(SparseVector(2, {0: 1.0})) == model.predict(SparseVector(2, {0: 1.0}))
    True
    >>> from shutil import rmtree
    >>> try:
    ...     rmtree(path)
    ... except OSError:
    ...     pass
    """
    def __init__(self, labels, pi, theta):
        self.labels = labels
        self.pi = pi
        self.theta = theta

    @since('0.9.0')
    def predict(self, x):
        """
        Return the most likely class for a data vector
        or an RDD of vectors
        """
        if isinstance(x, RDD):
            return x.map(lambda v: self.predict(v))
        x = _convert_to_vector(x)
        return self.labels[numpy.argmax(self.pi + x.dot(self.theta.transpose()))]

    def save(self, sc, path):
        """
        Save this model to the given path.
        """
        java_labels = _py2java(sc, self.labels.tolist())
        java_pi = _py2java(sc, self.pi.tolist())
        java_theta = _py2java(sc, self.theta.tolist())
        java_model = sc._jvm.org.apache.spark.mllib.classification.NaiveBayesModel(
            java_labels, java_pi, java_theta)
        java_model.save(sc._jsc.sc(), path)

    @classmethod
    @since('1.4.0')
    def load(cls, sc, path):
        """
        Load a model from the given path.
        """
        java_model = sc._jvm.org.apache.spark.mllib.classification.NaiveBayesModel.load(
            sc._jsc.sc(), path)
        # Can not unpickle array.array from Pyrolite in Python3 with "bytes"
        py_labels = _java2py(sc, java_model.labels(), "latin1")
        py_pi = _java2py(sc, java_model.pi(), "latin1")
        py_theta = _java2py(sc, java_model.theta(), "latin1")
        return NaiveBayesModel(py_labels, py_pi, numpy.array(py_theta))


class NaiveBayes(object):
    """
    Train a Multinomial Naive Bayes model.

    .. versionadded:: 0.9.0
    """

    @classmethod
    def train(cls, data, lambda_=1.0):
        """
        Train a Naive Bayes model given an RDD of (label, features)
        vectors.

        This is the `Multinomial NB <http://tinyurl.com/lsdw6p>`_ which
        can handle all kinds of discrete data.  For example, by
        converting documents into TF-IDF vectors, it can be used for
        document classification. By making every vector a 0-1 vector,
        it can also be used as `Bernoulli NB <http://tinyurl.com/p7c96j6>`_.
        The input feature values must be nonnegative.

        .. versionadded:: 0.9.0

        Parameters
        ----------
        data : :py:class:`pyspark.RDD`
            The training data, an RDD of :py:class:`pyspark.mllib.regression.LabeledPoint`.
        lambda\\_ : float, optional
            The smoothing parameter.
            (default: 1.0)
        """
        first = data.first()
        if not isinstance(first, LabeledPoint):
            raise ValueError("`data` should be an RDD of LabeledPoint")
        labels, pi, theta = callMLlibFunc("trainNaiveBayesModel", data, lambda_)
        return NaiveBayesModel(labels.toArray(), pi.toArray(), numpy.array(theta))


@inherit_doc
class StreamingLogisticRegressionWithSGD(StreamingLinearAlgorithm):
    """
    Train or predict a logistic regression model on streaming data.
    Training uses Stochastic Gradient Descent to update the model based on
    each new batch of incoming data from a DStream.

    Each batch of data is assumed to be an RDD of LabeledPoints.
    The number of data points per batch can vary, but the number
    of features must be constant. An initial weight
    vector must be provided.

    .. versionadded:: 1.5.0

    Parameters
    ----------
    stepSize : float, optional
        Step size for each iteration of gradient descent.
        (default: 0.1)
    numIterations : int, optional
        Number of iterations run for each batch of data.
        (default: 50)
    miniBatchFraction : float, optional
        Fraction of each batch of data to use for updates.
        (default: 1.0)
    regParam : float, optional
        L2 Regularization parameter.
        (default: 0.0)
    convergenceTol : float, optional
        Value used to determine when to terminate iterations.
        (default: 0.001)
    """
    def __init__(self, stepSize=0.1, numIterations=50, miniBatchFraction=1.0, regParam=0.0,
                 convergenceTol=0.001):
        self.stepSize = stepSize
        self.numIterations = numIterations
        self.regParam = regParam
        self.miniBatchFraction = miniBatchFraction
        self.convergenceTol = convergenceTol
        self._model = None
        super(StreamingLogisticRegressionWithSGD, self).__init__(
            model=self._model)

    @since('1.5.0')
    def setInitialWeights(self, initialWeights):
        """
        Set the initial value of weights.

        This must be set before running trainOn and predictOn.
        """
        initialWeights = _convert_to_vector(initialWeights)

        # LogisticRegressionWithSGD does only binary classification.
        self._model = LogisticRegressionModel(
            initialWeights, 0, initialWeights.size, 2)
        return self

    @since('1.5.0')
    def trainOn(self, dstream):
        """Train the model on the incoming dstream."""
        self._validate(dstream)

        def update(rdd):
            # LogisticRegressionWithSGD.train raises an error for an empty RDD.
            if not rdd.isEmpty():
                self._model = LogisticRegressionWithSGD.train(
                    rdd, self.numIterations, self.stepSize,
                    self.miniBatchFraction, self._model.weights,
                    regParam=self.regParam, convergenceTol=self.convergenceTol)

        dstream.foreachRDD(update)


def _test():
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.mllib.classification
    globs = pyspark.mllib.classification.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("mllib.classification tests")\
        .getOrCreate()
    globs['sc'] = spark.sparkContext
    (failure_count, test_count) = doctest.testmod(globs=globs, optionflags=doctest.ELLIPSIS)
    spark.stop()
    if failure_count:
        sys.exit(-1)

if __name__ == "__main__":
    _test()
